"""
Module printing summary of the contract
"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.myprettytable import MyPrettyTable
from slither.utils.output import Output


class VariableOrder(AbstractPrinter):
    ARGUMENT = "variable-order"
    HELP = "Print the storage order of the state variables"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#variable-order"

    def output(self, _filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""

        all_tables = []

        for contract in self.slither.contracts_derived:
            txt += f"\n{contract.name}:\n"
            table = MyPrettyTable(["Name", "Type", "Slot", "Offset", "State"])
            for variable in contract.storage_variables_ordered:
                slot, offset = contract.compilation_unit.storage_layout_of(contract, variable)
                table.add_row(
                    [variable.canonical_name, str(variable.type), slot, offset, "Storage"]
                )
            for variable in contract.transient_variables_ordered:
                slot, offset = contract.compilation_unit.storage_layout_of(contract, variable)
                table.add_row(
                    [variable.canonical_name, str(variable.type), slot, offset, "Transient"]
                )

            all_tables.append((contract.name, table))
            txt += str(table) + "\n"

        self.info(txt)

        res = self.generate_output(txt)
        for name, table in all_tables:
            res.add_pretty_table(table, name)
        return res



def unroll(list_to_unroll: list) -> list:
    ret = []
    for x in list_to_unroll:
        if not isinstance(x, list):
            ret += [x]
        else:
            ret += unroll(x)
    return ret



from slither.core.declarations import Contract
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.core.variables.variable import Variable
from slither.core.declarations.solidity_variables import SolidityVariable
from slither.core.declarations.function import Function

from slither.slithir.utils.utils import is_valid_lvalue
from slither.slithir.variables.constant import Constant
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA
from slither.slithir.variables.tuple import TupleVariable


class HighLevelCall(Call, OperationWithLValue):
    """
    High level message call
    """

    def __init__(
        self,
        destination: SourceMapping,
        function_name: Constant,
        nbr_arguments: int,
        result: TemporaryVariable | TupleVariable | TemporaryVariableSSA | None,
        type_call: str,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        assert isinstance(function_name, Constant)
        assert is_valid_lvalue(result) or result is None
        self._check_destination(destination)
        super().__init__(names=names)
        # Contract is only possible for library call, which inherits from highlevelcall
        self._destination: Variable | SolidityVariable | Contract = destination  # type: ignore
        self._function_name = function_name
        self._nbr_arguments = nbr_arguments
        self._type_call = type_call
        self._lvalue = result
        self._callid = None  # only used if gas/value != 0
        self._function_instance = None

        self._call_value = None
        self._call_gas = None

    # Development function, to be removed once the code is stable
    # It is overridden by LibraryCall
    def _check_destination(self, destination: Variable | SolidityVariable | Contract) -> None:
        assert isinstance(destination, (Variable, SolidityVariable))

    @property
    def call_id(self):
        return self._callid

    @call_id.setter
    def call_id(self, c):
        self._callid = c

    @property
    def call_value(self):
        return self._call_value

    @call_value.setter
    def call_value(self, v):
        self._call_value = v

    @property
    def call_gas(self):
        return self._call_gas

    @call_gas.setter
    def call_gas(self, v):
        self._call_gas = v

    @property
    def read(self) -> list[SourceMapping]:
        all_read = [self.destination, self.call_gas, self.call_value] + self._unroll(self.arguments)
        # remove None
        return [x for x in all_read if x]

    @property
    def destination(self) -> Variable | SolidityVariable | Contract:
        """
        Return a variable or a solidityVariable
        Contract is only possible for LibraryCall

        Returns:

        """
        return self._destination

    @property
    def function_name(self) -> Constant:
        return self._function_name

    @property
    def function(self) -> Function | Variable:
        return self._function_instance

    @function.setter
    def function(self, function):
        self._function_instance = function

    @property
    def nbr_arguments(self) -> int:
        return self._nbr_arguments

    @property
    def type_call(self) -> str:
        return self._type_call

    ###################################################################################
    ###################################################################################
    # region Analyses
    ###################################################################################
    ###################################################################################
    def is_static_call(self) -> bool:
        # If solidity >0.5, STATICCALL is used
        if self.compilation_unit.solc_version and self.compilation_unit.solc_version >= "0.5.0":
            if isinstance(self.function, Function) and (self.function.view or self.function.pure):
                return True
            if isinstance(self.function, Variable):
                return True
        return False

    def can_reenter(self, callstack: list[Function | Variable] | None = None) -> bool:
        """
        Must be called after slithIR analysis pass
        For Solidity > 0.5, filter access to public variables and constant/pure/view
        For call to this. check if the destination can re-enter
        :param callstack: check for recursion
        :return: bool
        """
        if self.is_static_call():
            return False
        # If there is a call to itself
        # We can check that the function called is
        # reentrancy-safe
        if self.destination == SolidityVariable("this"):
            if isinstance(self.function, Variable):
                return False
            # In case of recursion, return False
            callstack = [] if callstack is None else callstack
            if self.function in callstack:
                return False
            callstack = callstack + [self.function]
            if self.function.can_reenter(callstack):
                return True
        if isinstance(self.destination, Variable):
            if not self.destination.is_reentrant:
                return False

        return True

    def can_send_eth(self) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        return self._call_value is not None

    # endregion
    ###################################################################################
    ###################################################################################
    # region Built in
    ###################################################################################
    ###################################################################################

    def __str__(self):
        value = ""
        gas = ""
        if self.call_value:
            value = f"value:{self.call_value}"
        if self.call_gas:
            gas = f"gas:{self.call_gas}"
        arguments = []
        if self.arguments:
            arguments = self.arguments

        txt = "{}HIGH_LEVEL_CALL, dest:{}({}), function:{}, arguments:{} {} {}"
        if not self.lvalue:
            lvalue = ""
        elif isinstance(self.lvalue.type, (list,)):
            lvalue = f"{self.lvalue}({','.join(str(x) for x in self.lvalue.type)}) = "
        else:
            lvalue = f"{self.lvalue}({self.lvalue.type}) = "
        return txt.format(
            lvalue,
            self.destination,
            self.destination.type,
            self.function_name,
            [str(x) for x in arguments],
            value,
            gas,
        )



import os
from pathlib import Path
import sys

import pytest
from crytic_compile import CryticCompile, save_to_zip
from crytic_compile.utils.zip import load_from_zip

from solc_select import solc_select

from slither import Slither
from slither.detectors.abstract_detector import AbstractDetector
from slither.detectors import all_detectors


class Test:
    def __init__(
        self,
        detector: type[AbstractDetector],
        test_file: str,
        solc_ver: str,
        additional_files: list[str] | None = None,
    ):
        """


        :param detector:
        :param test_file:
        :param solc_ver:
        :param additional_files: If the test changes additional files, list them here to allow the
        test to update the source mapping
        """
        self.detector = detector
        self.test_file = test_file
        self.solc_ver = solc_ver
        if additional_files is None:
            self.additional_files = []
        else:
            self.additional_files = additional_files


def set_solc(test_item: Test):
    # hacky hack hack to pick the solc version we want
    env = dict(os.environ)

    if not solc_select.artifact_path(test_item.solc_ver).exists():
        print("Installing solc version", test_item.solc_ver)
        solc_select.install_artifacts([test_item.solc_ver])
    env["SOLC_VERSION"] = test_item.solc_ver
    os.environ.clear()
    os.environ.update(env)


def id_test(test_item: Test):
    return f"{test_item.detector.__name__}-{test_item.solc_ver}-{test_item.test_file}"


ALL_TESTS = [
    Test(
        all_detectors.UninitializedFunctionPtrsConstructor,
        "uninitialized_function_ptr_constructor.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UninitializedFunctionPtrsConstructor,
        "uninitialized_function_ptr_constructor.sol",
        "0.5.8",
    ),
    Test(
        all_detectors.UninitializedFunctionPtrsConstructor,
        "uninitialized_function_ptr_constructor.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyBenign,
        "reentrancy-benign.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ReentrancyBenign,
        "reentrancy-benign.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyBenign,
        "reentrancy-benign.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ReentrancyBenign,
        "reentrancy-benign.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "reentrancy-write.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "reentrancy-write.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "reentrancy-write.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "reentrancy-write.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "DAO.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "comment.sol",
        "0.8.2",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "no-reentrancy-staticcall.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "no-reentrancy-staticcall.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ReentrancyReadBeforeWritten,
        "no-reentrancy-staticcall.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.BooleanEquality,
        "boolean-constant-equality.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.BooleanEquality,
        "boolean-constant-equality.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.BooleanEquality,
        "boolean-constant-equality.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.BooleanEquality,
        "boolean-constant-equality.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.BooleanConstantMisuse,
        "boolean-constant-misuse.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.BooleanConstantMisuse,
        "boolean-constant-misuse.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.BooleanConstantMisuse,
        "boolean-constant-misuse.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.BooleanConstantMisuse,
        "boolean-constant-misuse.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UncheckedLowLevel,
        "unchecked_lowlevel.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UncheckedLowLevel,
        "unchecked_lowlevel.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UncheckedLowLevel,
        "unchecked_lowlevel.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UncheckedLowLevel,
        "unchecked_lowlevel.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnindexedERC20EventParameters,
        "erc20_indexed.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnindexedERC20EventParameters,
        "erc20_indexed.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnindexedERC20EventParameters,
        "erc20_indexed.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnindexedERC20EventParameters,
        "erc20_indexed.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.IncorrectERC20InterfaceDetection,
        "incorrect_erc20_interface.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.IncorrectERC20InterfaceDetection,
        "incorrect_erc20_interface.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.IncorrectERC20InterfaceDetection,
        "incorrect_erc20_interface.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.IncorrectERC20InterfaceDetection,
        "incorrect_erc20_interface.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.IncorrectERC721InterfaceDetection,
        "incorrect_erc721_interface.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.IncorrectERC721InterfaceDetection,
        "incorrect_erc721_interface.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.IncorrectERC721InterfaceDetection,
        "incorrect_erc721_interface.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.IncorrectERC721InterfaceDetection,
        "incorrect_erc721_interface.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UninitializedStateVarsDetection,
        "uninitialized.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UninitializedStateVarsDetection,
        "uninitialized.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UninitializedStateVarsDetection,
        "uninitialized.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UninitializedStateVarsDetection,
        "uninitialized.sol",
        "0.7.6",
    ),
    Test(all_detectors.Backdoor, "backdoor.sol", "0.4.25"),
    Test(all_detectors.Backdoor, "backdoor.sol", "0.5.16"),
    Test(all_detectors.Backdoor, "backdoor.sol", "0.6.11"),
    Test(all_detectors.Backdoor, "backdoor.sol", "0.7.6"),
    Test(all_detectors.Suicidal, "suicidal.sol", "0.4.25"),
    Test(all_detectors.Suicidal, "suicidal.sol", "0.5.16"),
    Test(all_detectors.Suicidal, "suicidal.sol", "0.6.11"),
    Test(all_detectors.Suicidal, "suicidal.sol", "0.7.6"),
    Test(
        all_detectors.ConstantPragma,
        "pragma.0.4.25.sol",
        "0.4.25",
        ["pragma.0.4.24.sol"],
    ),
    Test(
        all_detectors.ConstantPragma,
        "pragma.0.5.16.sol",
        "0.5.16",
        ["pragma.0.5.15.sol"],
    ),
    Test(
        all_detectors.ConstantPragma,
        "pragma.0.6.11.sol",
        "0.6.11",
        ["pragma.0.6.10.sol"],
    ),
    Test(
        all_detectors.ConstantPragma,
        "pragma.0.7.6.sol",
        "0.7.6",
        ["pragma.0.7.5.sol"],
    ),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.4.25"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.5.14"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.5.16"),
    Test(all_detectors.IncorrectSolc, "dynamic_1.sol", "0.5.16"),
    Test(all_detectors.IncorrectSolc, "dynamic_2.sol", "0.5.16"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.6.10"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.6.11"),
    Test(all_detectors.IncorrectSolc, "dynamic_1.sol", "0.6.11"),
    Test(all_detectors.IncorrectSolc, "dynamic_2.sol", "0.6.11"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.7.4"),
    Test(all_detectors.IncorrectSolc, "static.sol", "0.7.6"),
    Test(all_detectors.IncorrectSolc, "dynamic_1.sol", "0.7.6"),
    Test(all_detectors.IncorrectSolc, "dynamic_2.sol", "0.7.6"),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy_indirect.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy_indirect.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy_indirect.sol",
        "0.6.11",
    ),
    Test(all_detectors.ReentrancyEth, "reentrancy.sol", "0.7.6"),
    Test(
        all_detectors.ReentrancyEth,
        "reentrancy_indirect.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReentrancyEth,
        "DAO.sol",
        "0.4.25",
    ),
    # Test the nonReentrant filtering
    Test(all_detectors.ReentrancyEth, "reentrancy_with_non_reentrant.sol", "0.8.10"),
    # Test parse_ignore_comments
    Test(all_detectors.ReentrancyEth, "reentrancy_filtered_comments.sol", "0.8.10"),
    Test(
        all_detectors.UninitializedStorageVars,
        "uninitialized_storage_pointer.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UninitializedStorageVars,
        "uninitialized_storage_pointer.sol",
        "0.8.19",
    ),
    Test(all_detectors.TxOrigin, "tx_origin.sol", "0.4.25"),
    Test(all_detectors.TxOrigin, "tx_origin.sol", "0.5.16"),
    Test(all_detectors.TxOrigin, "tx_origin.sol", "0.6.11"),
    Test(all_detectors.TxOrigin, "tx_origin.sol", "0.7.6"),
    Test(
        all_detectors.UnusedStateVars,
        "unused_state.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnusedStateVars,
        "unused_state.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnusedStateVars,
        "unused_state.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnusedStateVars,
        "unused_state.sol",
        "0.7.6",
    ),
    Test(all_detectors.LockedEther, "locked_ether.sol", "0.4.25"),
    Test(all_detectors.LockedEther, "locked_ether.sol", "0.5.16"),
    Test(all_detectors.LockedEther, "locked_ether.sol", "0.6.11"),
    Test(all_detectors.LockedEther, "locked_ether.sol", "0.7.6"),
    Test(
        all_detectors.ArbitrarySendEth,
        "arbitrary_send_eth.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ArbitrarySendEth,
        "arbitrary_send_eth.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ArbitrarySendEth,
        "arbitrary_send_eth.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ArbitrarySendEth,
        "arbitrary_send_eth.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_contract.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_library.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_contract.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_library.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_contract.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_library.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_contract.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.Assembly,
        "inline_assembly_library.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.LowLevelCalls,
        "low_level_calls.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.LowLevelCalls,
        "low_level_calls.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.LowLevelCalls,
        "low_level_calls.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.LowLevelCalls,
        "low_level_calls.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "const_state_variables.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "const_state_variables.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "const_state_variables.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "const_state_variables.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "const_state_variables.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.CouldBeConstant,
        "unused_yul.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.CouldBeImmutable,
        "immut_state_variables.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.CouldBeImmutable,
        "immut_state_variables.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.CouldBeImmutable,
        "immut_state_variables.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.CouldBeImmutable,
        "immut_state_variables.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.CouldBeImmutable,
        "immut_state_variables.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_2.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_3.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_2.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_3.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_2.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_3.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_2.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ExternalFunction,
        "external_function_3.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.NamingConvention,
        "naming_convention.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.NamingConvention,
        "naming_convention.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.NamingConvention,
        "naming_convention.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.NamingConvention,
        "naming_convention.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.NamingConvention,
        "no_warning_for_public_constants.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.NamingConvention,
        "no_warning_for_public_constants.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.NamingConvention,
        "no_warning_for_public_constants.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.NamingConvention,
        "no_warning_for_public_constants.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ControlledDelegateCall,
        "controlled_delegatecall.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ControlledDelegateCall,
        "controlled_delegatecall.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ControlledDelegateCall,
        "controlled_delegatecall.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ControlledDelegateCall,
        "controlled_delegatecall.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UninitializedLocalVars,
        "uninitialized_local_variable.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UninitializedLocalVars,
        "uninitialized_local_variable.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UninitializedLocalVars,
        "uninitialized_local_variable.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UninitializedLocalVars,
        "uninitialized_local_variable.sol",
        "0.7.6",
    ),
    Test(all_detectors.ConstantFunctionsAsm, "constant.sol", "0.4.25"),
    Test(
        all_detectors.ConstantFunctionsState,
        "constant.sol",
        "0.4.25",
    ),
    Test(all_detectors.ConstantFunctionsAsm, "constant.sol", "0.5.16"),
    Test(
        all_detectors.ConstantFunctionsState,
        "constant.sol",
        "0.5.16",
    ),
    Test(all_detectors.ConstantFunctionsAsm, "constant.sol", "0.6.11"),
    Test(
        all_detectors.ConstantFunctionsState,
        "constant.sol",
        "0.6.11",
    ),
    Test(all_detectors.ConstantFunctionsAsm, "constant.sol", "0.7.6"),
    Test(all_detectors.ConstantFunctionsState, "constant.sol", "0.7.6"),
    Test(
        all_detectors.UnusedReturnValues,
        "unused_return.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnusedReturnValues,
        "unused_return.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnusedReturnValues,
        "unused_return.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnusedReturnValues,
        "unused_return.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UncheckedTransfer,
        "unused_return_transfers.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ShadowingAbstractDetection,
        "shadowing_abstract.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ShadowingAbstractDetection,
        "shadowing_abstract.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ShadowingAbstractDetection,
        "shadowing_state_variable.sol",
        "0.7.5",
    ),
    Test(
        all_detectors.ShadowingAbstractDetection,
        "public_gap_variable.sol",
        "0.7.5",
    ),
    Test(
        all_detectors.StateShadowing,
        "shadowing_state_variable.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.StateShadowing,
        "shadowing_state_variable.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.StateShadowing,
        "shadowing_state_variable.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.StateShadowing,
        "shadowing_state_variable.sol",
        "0.7.5",
    ),
    Test(
        all_detectors.StateShadowing,
        "public_gap_variable.sol",
        "0.7.5",
    ),
    Test(
        all_detectors.StateShadowing,
        "shadowing_state_variable.sol",
        "0.7.6",
    ),
    Test(all_detectors.Timestamp, "timestamp.sol", "0.4.25"),
    Test(all_detectors.Timestamp, "timestamp.sol", "0.5.16"),
    Test(all_detectors.Timestamp, "timestamp.sol", "0.6.11"),
    Test(all_detectors.Timestamp, "timestamp.sol", "0.7.6"),
    Test(
        all_detectors.MultipleCallsInLoop,
        "multiple_calls_in_loop.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MultipleCallsInLoop,
        "multiple_calls_in_loop.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MultipleCallsInLoop,
        "multiple_calls_in_loop.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MultipleCallsInLoop,
        "multiple_calls_in_loop.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.BuiltinSymbolShadowing,
        "shadowing_builtin_symbols.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.BuiltinSymbolShadowing,
        "shadowing_builtin_symbols.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.LocalShadowing,
        "shadowing_local_variable.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.LocalShadowing,
        "shadowing_local_variable.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.LocalShadowing,
        "shadowing_local_variable.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.LocalShadowing,
        "shadowing_local_variable.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.RightToLeftOverride,
        "right_to_left_override.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.RightToLeftOverride,
        "right_to_left_override.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.RightToLeftOverride,
        "right_to_left_override.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.RightToLeftOverride,
        "unicode_direction_override.sol",
        "0.8.0",
    ),
    Test(all_detectors.VoidConstructor, "void-cst.sol", "0.4.25"),
    Test(all_detectors.VoidConstructor, "void-cst.sol", "0.5.16"),
    Test(all_detectors.VoidConstructor, "void-cst.sol", "0.6.11"),
    Test(all_detectors.VoidConstructor, "void-cst.sol", "0.7.6"),
    Test(
        all_detectors.UncheckedSend,
        "unchecked_send.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UncheckedSend,
        "unchecked_send.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UncheckedSend,
        "unchecked_send.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UncheckedSend,
        "unchecked_send.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReentrancyEvent,
        "reentrancy-events.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ReentrancyEvent,
        "reentrancy-events.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ReentrancyEvent,
        "reentrancy-events.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReentrancyBalance,
        "reentrancy_balance.sol",
        "0.8.10",
    ),
    Test(
        all_detectors.IncorrectStrictEquality,
        "incorrect_equality.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.IncorrectStrictEquality,
        "incorrect_equality.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.IncorrectStrictEquality,
        "incorrect_equality.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.IncorrectStrictEquality,
        "incorrect_equality.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.TooManyDigits,
        "too_many_digits.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.TooManyDigits,
        "too_many_digits.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.TooManyDigits,
        "too_many_digits.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.TooManyDigits,
        "too_many_digits.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Buggy.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Fixed.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "whitelisted.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Reinitializer.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "AnyInitializer.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Buggy.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Fixed.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "whitelisted.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Reinitializer.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "AnyInitializer.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Buggy.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Fixed.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "whitelisted.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Reinitializer.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "AnyInitializer.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Buggy.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Fixed.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Reinitializer.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "AnyInitializer.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "whitelisted.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Buggy.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Fixed.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "whitelisted.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "Reinitializer.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.UnprotectedUpgradeable,
        "AnyInitializer.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.ABIEncoderV2Array,
        "storage_ABIEncoderV2_array.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ABIEncoderV2Array,
        "storage_ABIEncoderV2_array.sol",
        "0.5.10",
    ),
    Test(
        all_detectors.ABIEncoderV2Array,
        "storage_ABIEncoderV2_array.sol",
        "0.5.9",
    ),
    Test(
        all_detectors.ArrayByReference,
        "array_by_reference.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ArrayByReference,
        "array_by_reference.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ArrayByReference,
        "array_by_reference.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ArrayByReference,
        "array_by_reference.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.AssertStateChange,
        "assert_state_change.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.AssertStateChange,
        "assert_state_change.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.AssertStateChange,
        "assert_state_change.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.AssertStateChange,
        "assert_state_change.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ArrayLengthAssignment,
        "array_length_assignment.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ArrayLengthAssignment,
        "array_length_assignment.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.CostlyOperationsInLoop,
        "multiple_costly_operations_in_loop.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.CostlyOperationsInLoop,
        "multiple_costly_operations_in_loop.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.CostlyOperationsInLoop,
        "multiple_costly_operations_in_loop.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.CostlyOperationsInLoop,
        "multiple_costly_operations_in_loop.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.FunctionInitializedState,
        "function_init_state_variables.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.FunctionInitializedState,
        "function_init_state_variables.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.FunctionInitializedState,
        "function_init_state_variables.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.FunctionInitializedState,
        "function_init_state_variables.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.MappingDeletionDetection,
        "MappingDeletion.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MappingDeletionDetection,
        "MappingDeletion.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MappingDeletionDetection,
        "MappingDeletion.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MappingDeletionDetection,
        "MappingDeletion.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.PublicMappingNested,
        "public_mappings_nested.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.RedundantStatements,
        "redundant_statements.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.RedundantStatements,
        "redundant_statements.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.RedundantStatements,
        "redundant_statements.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.RedundantStatements,
        "redundant_statements.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ReusedBaseConstructor,
        "reused_base_constructor.sol",
        "0.4.21",
    ),
    Test(
        all_detectors.ReusedBaseConstructor,
        "reused_base_constructor.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.StorageSignedIntegerArray,
        "storage_signed_integer_array.sol",
        "0.5.10",
    ),
    Test(
        all_detectors.StorageSignedIntegerArray,
        "storage_signed_integer_array.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented_interfaces.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented_interfaces.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.UnimplementedFunctionDetection,
        "unimplemented_interfaces.sol",
        "0.7.6",
    ),
    Test(all_detectors.BadPRNG, "bad_prng.sol", "0.4.25"),
    Test(all_detectors.BadPRNG, "bad_prng.sol", "0.5.16"),
    Test(all_detectors.BadPRNG, "bad_prng.sol", "0.6.11"),
    Test(all_detectors.BadPRNG, "bad_prng.sol", "0.7.6"),
    Test(
        all_detectors.MissingEventsAccessControl,
        "missing_events_access_control.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MissingEventsAccessControl,
        "missing_events_access_control.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MissingEventsAccessControl,
        "missing_events_access_control.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MissingEventsAccessControl,
        "missing_events_access_control.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.MissingEventsArithmetic,
        "missing_events_arithmetic.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MissingEventsArithmetic,
        "missing_events_arithmetic.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MissingEventsArithmetic,
        "missing_events_arithmetic.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MissingEventsArithmetic,
        "missing_events_arithmetic.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ModifierDefaultDetection,
        "modifier_default.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ModifierDefaultDetection,
        "modifier_default.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ModifierDefaultDetection,
        "modifier_default.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ModifierDefaultDetection,
        "modifier_default.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.IncorrectUnaryExpressionDetection,
        "invalid_unary_expression.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MissingZeroAddressValidation,
        "missing_zero_address_validation.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MissingZeroAddressValidation,
        "missing_zero_address_validation.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MissingZeroAddressValidation,
        "missing_zero_address_validation.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MissingZeroAddressValidation,
        "missing_zero_address_validation.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.PredeclarationUsageLocal,
        "predeclaration_usage_local.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DeadCode,
        "dead-code.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.WriteAfterWrite,
        "write-after-write.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.ShiftParameterMixup,
        "shift_parameter_mixup.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ShiftParameterMixup,
        "shift_parameter_mixup.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ShiftParameterMixup,
        "shift_parameter_mixup.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ShiftParameterMixup,
        "shift_parameter_mixup.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.MissingInheritance,
        "unimplemented_interface.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MissingInheritance,
        "unimplemented_interface.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MissingInheritance,
        "unimplemented_interface.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MissingInheritance,
        "unimplemented_interface.sol",
        "0.7.6",
    ),
    # Does not work on the CI. Most likely because of solc 0.4.2?
    # Test(
    #     all_detectors.EnumConversion,
    #     "enum_conversion.sol",
    #     "0.4.2",
    # ),
    Test(
        all_detectors.MultipleConstructorSchemes,
        "multiple_constructor_schemes.sol",
        "0.4.22",
    ),
    Test(
        all_detectors.DeprecatedStandards,
        "deprecated_calls.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DivideBeforeMultiply,
        "divide_before_multiply.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DivideBeforeMultiply,
        "divide_before_multiply.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.DivideBeforeMultiply,
        "divide_before_multiply.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.DivideBeforeMultiply,
        "divide_before_multiply.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.TypeBasedTautology,
        "type_based_tautology.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.TypeBasedTautology,
        "type_based_tautology.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.TypeBasedTautology,
        "type_based_tautology.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.TypeBasedTautology,
        "type_based_tautology.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.MsgValueInLoop,
        "msg_value_loop.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.MsgValueInLoop,
        "msg_value_loop.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.MsgValueInLoop,
        "msg_value_loop.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.MsgValueInLoop,
        "msg_value_loop.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.MsgValueInLoop,
        "msg_value_loop.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.DelegatecallInLoop,
        "delegatecall_loop.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DelegatecallInLoop,
        "delegatecall_loop.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.DelegatecallInLoop,
        "delegatecall_loop.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.DelegatecallInLoop,
        "delegatecall_loop.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.DelegatecallInLoop,
        "delegatecall_loop.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.ProtectedVariables,
        "comment.sol",
        "0.8.2",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.ArbitrarySendErc20NoPermit,
        "arbitrary_send_erc20_inheritance.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.ArbitrarySendErc20Permit,
        "arbitrary_send_erc20_permit.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.ArbitrarySendErc20Permit,
        "arbitrary_send_erc20_permit.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.ArbitrarySendErc20Permit,
        "arbitrary_send_erc20_permit.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.ArbitrarySendErc20Permit,
        "arbitrary_send_erc20_permit.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.ArbitrarySendErc20Permit,
        "arbitrary_send_erc20_permit.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_collision.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_collision.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_collision.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_collision.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_collision.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_wrong_return_type.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_wrong_return_type.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_wrong_return_type.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_wrong_return_type.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_wrong_return_type.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_state_var_collision.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_state_var_collision.sol",
        "0.5.16",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_state_var_collision.sol",
        "0.6.11",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_state_var_collision.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.DomainSeparatorCollision,
        "permit_domain_state_var_collision.sol",
        "0.8.0",
    ),
    Test(
        all_detectors.VarReadUsingThis,
        "var_read_using_this.sol",
        "0.4.25",
    ),
    Test(
        all_detectors.VarReadUsingThis,
        "var_read_using_this.sol",
        "0.5.16",
    ),
    Test(all_detectors.VarReadUsingThis, "var_read_using_this.sol", "0.6.11"),
    Test(
        all_detectors.VarReadUsingThis,
        "var_read_using_this.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.VarReadUsingThis,
        "var_read_using_this.sol",
        "0.8.15",
    ),
    Test(
        all_detectors.CyclomaticComplexity,
        "HighCyclomaticComplexity.sol",
        "0.8.16",
    ),
    Test(
        all_detectors.CyclomaticComplexity,
        "LowCyclomaticComplexity.sol",
        "0.8.16",
    ),
    Test(
        all_detectors.CacheArrayLength,
        "CacheArrayLength.sol",
        "0.8.17",
    ),
    Test(
        all_detectors.IncorrectUsingFor,
        "IncorrectUsingForTopLevel.sol",
        "0.8.17",
    ),
    Test(
        all_detectors.EncodePackedCollision,
        "encode_packed_collision.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.IncorrectReturn,
        "incorrect_return.sol",
        "0.8.10",
    ),
    Test(
        all_detectors.ReturnInsteadOfLeave,
        "incorrect_return.sol",
        "0.8.10",
    ),
    Test(
        all_detectors.IncorrectOperatorExponentiation,
        "incorrect_exp.sol",
        "0.7.6",
    ),
    Test(
        all_detectors.TautologicalCompare,
        "compare.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.ReturnBomb,
        "return_bomb.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.OutOfOrderRetryable,
        "out_of_order_retryable.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.GelatoUnprotectedRandomness,
        "gelato_unprotected_randomness.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.ChronicleUncheckedPrice,
        "chronicle_unchecked_price.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.PythUncheckedConfidence,
        "pyth_unchecked_confidence.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.PythUncheckedPublishTime,
        "pyth_unchecked_publishtime.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.ChainlinkFeedRegistry,
        "chainlink_feed_registry.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.PythDeprecatedFunctions,
        "pyth_deprecated_functions.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.OptimismDeprecation,
        "optimism_deprecation.sol",
        "0.8.20",
    ),
    Test(
        all_detectors.UnindexedEventAddress,
        "unindexed_event_address.sol",
        "0.8.22",
    ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ConstantContractLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ConstantContractLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ConstantTopLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ConstantTopLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ContractUsedInContractTest1.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ContractUsedInContractTest2.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "ContractUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomErrorTopLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomEventContractLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomEventContractLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedInContractTest1.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedInContractTest2.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedInContractTest3.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedInContractTest4.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedTopLevelTest1.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeContractLevelUsedTopLevelTest2.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedInContractTest1.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedInContractTest2.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedInContractTest3.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedInContractTest4.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedTopLevelTest1.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "CustomTypeTopLevelUsedTopLevelTest2.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "EnumContractLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "EnumContractLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "EnumTopLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "EnumTopLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "FunctionContractLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "FunctionContractLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "FunctionTopLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "FunctionTopLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "LibraryUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "LibraryUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "StructContractLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "StructContractLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "StructTopLevelUsedInContractTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "StructTopLevelUsedTopLevelTest.sol",
    #     "0.8.16",
    # ),
    # Test(
    #     all_detectors.UnusedImport,
    #     "C.sol",
    #     "0.8.16",
    # ),
]

GENERIC_PATH = "/GENERIC_PATH"

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


@pytest.mark.parametrize("test_item", ALL_TESTS, ids=id_test)
def test_detector(test_item: Test, snapshot):
    test_dir_path = Path(
        TEST_DATA_DIR,
        test_item.detector.ARGUMENT,
        test_item.solc_ver,
    ).as_posix()
    test_file_path = Path(test_dir_path, test_item.test_file).as_posix()

    zip_artifact_path = Path(f"{test_file_path}-{test_item.solc_ver}.zip").as_posix()
    crytic_compile = load_from_zip(zip_artifact_path)[0]

    sl = Slither(crytic_compile)
    sl.register_detector(test_item.detector)
    results = sl.run_detectors()

    actual_output = ""
    for detector_result in results:
        for result in detector_result:
            actual_output += result["description"]
            actual_output += "\n"
    assert snapshot() == actual_output


def _generate_compile(test_item: Test, skip_existing=False):
    test_dir_path = Path(
        TEST_DATA_DIR,
        test_item.detector.ARGUMENT,
        test_item.solc_ver,
    ).as_posix()
    test_file_path = Path(test_dir_path, test_item.test_file).as_posix()
    zip_artifact_path = Path(f"{test_file_path}-{test_item.solc_ver}.zip").as_posix()

    if skip_existing:
        if os.path.isfile(zip_artifact_path):
            return

    set_solc(test_item)
    crytic_compile = CryticCompile(test_file_path)
    save_to_zip([crytic_compile], zip_artifact_path)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(
            "To generate the zip artifacts run\n\tpython tests/e2e/tests/test_detectors.py --compile"
        )
    elif sys.argv[1] == "--compile":
        for next_test in ALL_TESTS:
            _generate_compile(next_test, skip_existing=True)



"""
Module detecting dead code
"""

from slither.core.declarations import Function, FunctionContract, Contract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class DeadCode(AbstractDetector):
    """
    Unprotected function detector
    """

    ARGUMENT = "dead-code"
    HELP = "Functions that are not used"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#dead-code"

    WIKI_TITLE = "Dead-code"
    WIKI_DESCRIPTION = "Functions that are not used."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    function dead_code() internal() {}
}
```
`dead_code` is not used in the contract, and make the code's review more difficult."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove unused functions."

    def _detect(self) -> list[Output]:
        results = []

        functions_used = set()
        for contract in self.compilation_unit.contracts_derived:
            all_functionss_called = [
                f.all_internal_calls() for f in contract.functions_entry_points
            ]
            all_functions_called = [
                item.function for sublist in all_functionss_called for item in sublist
            ]
            functions_used |= {
                f.canonical_name for f in all_functions_called if isinstance(f, Function)
            }
            all_libss_called = [f.all_library_calls() for f in contract.functions_entry_points]
            all_libs_called: list[tuple[Contract, Function]] = [
                item.function for sublist in all_libss_called for item in sublist
            ]
            functions_used |= {
                lib[1].canonical_name for lib in all_libs_called if isinstance(lib, tuple)
            }
        for function in sorted(self.compilation_unit.functions, key=lambda x: x.canonical_name):
            if (
                function.visibility in ["public", "external"]
                or function.is_constructor
                or function.is_fallback
                or function.is_constructor_variables
            ):
                continue
            if function.canonical_name in functions_used:
                continue
            if isinstance(function, FunctionContract) and (
                function.contract_declarer.is_from_dependency()
                or function.contract_declarer.is_library
            ):
                continue
            # Continue if the function is not implemented because it means the contract is abstract
            if not function.is_implemented:
                continue
            info: DETECTOR_INFO = [function, " is never used and should be removed\n"]
            res = self.generate_result(info)
            results.append(res)

        return results



import sys
import argparse
import logging
from crytic_compile import cryticparser
from slither import Slither
from slither.utils.command_line import read_config_file
from slither.tools.slither_format.slither_format import slither_format


logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)

# Slither detectors for which slither-format currently works
available_detectors = [
    "unused-state",
    "solc-version",
    "pragma",
    "naming-convention",
    "external-function",
    "constable-states",
    "constant-function-asm",
    "constatnt-function-state",
]


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(description="slither_format", usage="slither_format filename")

    parser.add_argument(
        "filename", help="The filename of the contract or truffle directory to analyze."
    )
    parser.add_argument(
        "--verbose-test",
        "-v",
        help="verbose mode output for testing",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--verbose-json",
        "-j",
        help="verbose json output",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        help="displays the current version",
        version="0.1.0",
        action="version",
    )

    parser.add_argument(
        "--config-file",
        help="Provide a config file (default: slither.config.json)",
        action="store",
        dest="config_file",
        default="slither.config.json",
    )

    group_detector = parser.add_argument_group("Detectors")
    group_detector.add_argument(
        "--detect",
        help="Comma-separated list of detectors, defaults to all, "
        f"available detectors: {', '.join(d for d in available_detectors)}",
        action="store",
        dest="detectors_to_run",
        default="all",
    )

    group_detector.add_argument(
        "--exclude",
        help="Comma-separated list of detectors to exclude,"
        "available detectors: {', '.join(d for d in available_detectors)}",
        action="store",
        dest="detectors_to_exclude",
        default="all",
    )

    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def main() -> None:
    # ------------------------------
    #       Usage: python3 -m slither_format filename
    #       Example: python3 -m slither_format contract.sol
    # ------------------------------
    # Parse all arguments
    args = parse_args()

    read_config_file(args)

    # Perform slither analysis on the given filename
    slither = Slither(args.filename, **vars(args))

    # Format the input files based on slither analysis
    slither_format(slither, **vars(args))


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.top_level import TopLevel

if TYPE_CHECKING:
    from slither.core.compilation_unit import SlitherCompilationUnit
    from slither.core.scope.scope import FileScope


class CustomErrorTopLevel(CustomError, TopLevel):
    def __init__(self, compilation_unit: "SlitherCompilationUnit", scope: "FileScope") -> None:
        super().__init__(compilation_unit)
        self.file_scope: FileScope = scope

    @property
    def canonical_name(self) -> str:
        return self.full_name



import sys
from slither.slither import Slither


if len(sys.argv) != 2:
    print("python function_called.py contract.sol")
    sys.exit(-1)

# Init slither
slither = Slither(sys.argv[1])

for contract in slither.contracts:
    for function in contract.functions + contract.modifiers:
        filename = f"{sys.argv[1]}-{contract.name}-{function.full_name}.dot"
        print(f"Export {filename}")
        function.slithir_cfg_to_dot(filename)



from typing import Union

from slither.core.expressions.expression import Expression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.literal import Literal


class IndexAccess(Expression):
    def __init__(
        self,
        left_expression: Union["IndexAccess", Identifier],
        right_expression: Literal | Identifier,
    ) -> None:
        super().__init__()
        self._expressions = [left_expression, right_expression]

    @property
    def expressions(self) -> list["Expression"]:
        return self._expressions

    @property
    def expression_left(self) -> "Expression":
        return self._expressions[0]

    @property
    def expression_right(self) -> "Expression":
        return self._expressions[1]

    def __str__(self) -> str:
        return str(self.expression_left) + "[" + str(self.expression_right) + "]"



from slither.detectors.abstract_detector import (
    AbstractDetector,
    DETECTOR_INFO,
)
from slither.utils.output import Output
from slither.slithir.operations import Member, Binary, Assignment


class PythUnchecked(AbstractDetector):
    """
    Documentation: This detector finds deprecated Pyth function calls
    """

    # To be overridden in the derived class
    PYTH_FUNCTIONS = []
    PYTH_FIELD = ""

    def _detect(self) -> list[Output]:
        results: list[Output] = []

        for contract in self.compilation_unit.contracts_derived:
            for target_contract, ir in contract.all_high_level_calls:
                if target_contract.name == "IPyth" and ir.function_name in self.PYTH_FUNCTIONS:
                    # We know for sure the last IR in the node is an Assignment operation of the TMP variable. Example:
                    # 	Expression: price = pyth.getEmaPriceNoOlderThan(id,age)
                    #   IRs:
                    #      TMP_0(PythStructs.Price) = HIGH_LEVEL_CALL, dest:pyth(IPyth), function:getEmaPriceNoOlderThan, arguments:['id', 'age']
                    #      price(PythStructs.Price) := TMP_0(PythStructs.Price)
                    assert isinstance(ir.node.irs[len(ir.node.irs) - 1], Assignment)
                    return_variable = ir.node.irs[len(ir.node.irs) - 1].lvalue
                    checked = False

                    possible_unchecked_variable_ir = None
                    nodes = ir.node.sons
                    visited = set()
                    while nodes:
                        if checked:
                            break
                        next_node = nodes[0]
                        nodes = nodes[1:]

                        for node_ir in next_node.all_slithir_operations():
                            # We are accessing the unchecked_var field of the returned Price struct
                            if (
                                isinstance(node_ir, Member)
                                and node_ir.variable_left == return_variable
                                and node_ir.variable_right.name == self.PYTH_FIELD
                            ):
                                possible_unchecked_variable_ir = node_ir.lvalue
                            # We assume that if unchecked_var happens to be inside a binary operation is checked
                            if (
                                isinstance(node_ir, Binary)
                                and possible_unchecked_variable_ir is not None
                                and possible_unchecked_variable_ir in node_ir.read
                            ):
                                checked = True
                                break

                        if next_node not in visited:
                            visited.add(next_node)
                            for son in next_node.sons:
                                if son not in visited:
                                    nodes.append(son)

                    if not checked:
                        info: DETECTOR_INFO = [
                            f"Pyth price {self.PYTH_FIELD} field is not checked in ",
                            ir.node.function,
                            "\n\t- ",
                            ir.node,
                            "\n",
                        ]
                        res = self.generate_result(info)
                        results.append(res)

        return results



from slither.core.declarations import Function
from slither.core.variables import Variable
from slither.slithir.operations.operation import Operation


class Call(Operation):
    def __init__(self, names: list[str] | None = None) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        assert (names is None) or isinstance(names, list)
        super().__init__()
        self._arguments: list[Variable] = []
        self._names = names

    @property
    def names(self) -> list[str] | None:
        """
        For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
        Otherwise, None.
        """
        return self._names

    @property
    def arguments(self) -> list[Variable]:
        return self._arguments

    @arguments.setter
    def arguments(self, v: list[Variable]) -> None:
        self._arguments = v

    def can_reenter(self, _callstack: list[Function | Variable] | None = None) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        return False

    def can_send_eth(self) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        return False



from slither.core.variables.variable import Variable


class SlithIRVariable(Variable):
    def __init__(self) -> None:
        super().__init__()
        self._index = 0

    @property
    def ssa_name(self) -> str:
        assert self.name
        return self.name

    def __str__(self) -> str:
        return self.ssa_name



"""
Detect incorrect erc20 interface.
Some contracts do not return a bool on transfer/transferFrom/approve, which may lead to preventing the contract to be used with contracts compiled with recent solc (>0.4.22)
"""

from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class IncorrectERC20InterfaceDetection(AbstractDetector):
    """
    Incorrect ERC20 Interface
    """

    ARGUMENT = "erc20-interface"
    HELP = "Incorrect ERC20 interfaces"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-erc20-interface"

    WIKI_TITLE = "Incorrect erc20 interface"
    WIKI_DESCRIPTION = "Incorrect return values for `ERC20` functions. A contract compiled with Solidity > 0.4.22 interacting with these functions will fail to execute them, as the return value is missing."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Token{
    function transfer(address to, uint value) external;
    //...
}
```
`Token.transfer` does not return a boolean. Bob deploys the token. Alice creates a contract that interacts with it but assumes a correct `ERC20` interface implementation. Alice's contract is unable to interact with Bob's contract."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Set the appropriate return values and types for the defined `ERC20` functions."
    )

    @staticmethod
    def incorrect_erc20_interface(signature: tuple[str, list[str], list[str]]) -> bool:
        (name, parameters, returnVars) = signature

        if name == "transfer" and parameters == ["address", "uint256"] and returnVars != ["bool"]:
            return True

        if (
            name == "transferFrom"
            and parameters == ["address", "address", "uint256"]
            and returnVars != ["bool"]
        ):
            return True

        if name == "approve" and parameters == ["address", "uint256"] and returnVars != ["bool"]:
            return True

        if (
            name == "allowance"
            and parameters == ["address", "address"]
            and returnVars != ["uint256"]
        ):
            return True

        if name == "balanceOf" and parameters == ["address"] and returnVars != ["uint256"]:
            return True

        if name == "totalSupply" and parameters == [] and returnVars != ["uint256"]:
            return True

        return False

    @staticmethod
    def detect_incorrect_erc20_interface(contract: Contract) -> list[FunctionContract]:
        """Detect incorrect ERC20 interface

        Returns:
            list(str) : list of incorrect function signatures
        """

        # Verify this is an ERC20 contract.
        if not contract.is_possible_erc20():
            return []

        # If this contract implements a function from ERC721, we can assume it is an ERC721 token. These tokens
        # offer functions which are similar to ERC20, but are not compatible.
        if contract.is_possible_erc721():
            return []

        funcs = contract.functions
        functions = [
            f
            for f in funcs
            if IncorrectERC20InterfaceDetection.incorrect_erc20_interface(f.signature)
        ]

        return functions

    def _detect(self) -> list[Output]:
        """Detect incorrect erc20 interface

        Returns:
            dict: [contract name] = set(str)  events
        """
        results = []
        for c in self.compilation_unit.contracts_derived:
            functions = IncorrectERC20InterfaceDetection.detect_incorrect_erc20_interface(c)
            if functions:
                for function in functions:
                    info: DETECTOR_INFO = [
                        c,
                        " has incorrect ERC20 function interface:",
                        function,
                        "\n",
                    ]
                    json = self.generate_result(info)

                    results.append(json)

        return results



import logging
from typing import Any
from collections.abc import Callable

import slither.slithir.variables.tuple_ssa
from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import (
    Contract,
    Enum,
    Function,
    SolidityFunction,
    SolidityVariable,
    Structure,
)
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.modifier import Modifier
from slither.core.declarations.solidity_import_placeholder import SolidityImportPlaceHolder
from slither.core.solidity_types.type import Type
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.top_level_variable import TopLevelVariable
from slither.core.variables.variable import Variable
from slither.slithir.exceptions import SlithIRError
from slither.slithir.operations import (
    Assignment,
    Binary,
    Condition,
    Delete,
    EventCall,
    HighLevelCall,
    Index,
    InitArray,
    InternalCall,
    InternalDynamicCall,
    Length,
    LibraryCall,
    LowLevelCall,
    Member,
    NewArray,
    NewContract,
    NewElementaryType,
    NewStructure,
    OperationWithLValue,
    Phi,
    PhiCallback,
    Return,
    Send,
    SolidityCall,
    Transfer,
    TypeConversion,
    Unary,
    Unpack,
    Nop,
)
from slither.slithir.operations.call import Call
from slither.slithir.operations.codesize import CodeSize
from slither.slithir.operations.operation import Operation
from slither.slithir.variables import (
    Constant,
    LocalIRVariable,
    ReferenceVariable,
    ReferenceVariableSSA,
    StateIRVariable,
    TemporaryVariable,
    TemporaryVariableSSA,
    TupleVariable,
    TupleVariableSSA,
)

logger = logging.getLogger("SSA_Conversion")

###################################################################################
###################################################################################
# region SlihtIR variables to SSA
###################################################################################
###################################################################################


def transform_slithir_vars_to_ssa(
    function: FunctionContract | Modifier | FunctionTopLevel,
) -> None:
    """
    Transform slithIR vars to SSA (TemporaryVariable, ReferenceVariable, TupleVariable)
    """
    variables = []
    for node in function.nodes:
        for ir in node.irs_ssa:
            if isinstance(ir, OperationWithLValue) and ir.lvalue not in variables:
                variables += [ir.lvalue]

    tmp_variables = [v for v in variables if isinstance(v, TemporaryVariable)]
    for idx, _ in enumerate(tmp_variables):
        tmp_variables[idx].index = idx
    ref_variables = [v for v in variables if isinstance(v, ReferenceVariable)]
    for idx, _ in enumerate(ref_variables):
        ref_variables[idx].index = idx
    tuple_variables = [v for v in variables if isinstance(v, TupleVariable)]
    for idx, _ in enumerate(tuple_variables):
        tuple_variables[idx].index = idx


###################################################################################
###################################################################################
# region SSA conversion
###################################################################################
###################################################################################


def add_ssa_ir(
    function: FunctionContract | Modifier | FunctionTopLevel,
    all_state_variables_instances: dict[
        str, slither.slithir.variables.state_variable.StateIRVariable
    ],
) -> None:
    """
        Add SSA version of the IR
    Args:
        function
        all_state_variables_instances
    """

    if not function.is_implemented:
        return

    init_definition = {}
    for v in function.parameters:
        if v.name:
            init_definition[v.name] = (v, function.entry_point)
            function.entry_point.add_ssa_ir(Phi(LocalIRVariable(v), set()))

    for v in function.returns:
        if v.name:
            init_definition[v.name] = (v, function.entry_point)

    # We only add phi function for state variable at entry node if
    # The state variable is used
    # And if the state variables is written in another function (otherwise its stay at index 0)
    for _, variable_instance in all_state_variables_instances.items():
        if is_used_later(function.entry_point, variable_instance):
            # rvalues are fixed in solc_parsing.declaration.function
            function.entry_point.add_ssa_ir(Phi(StateIRVariable(variable_instance), set()))

    add_phi_origins(function.entry_point, init_definition, {})

    for node in function.nodes:
        for variable, nodes in node.phi_origins_local_variables.values():
            if len(nodes) < 2:
                continue
            if not is_used_later(node, variable):
                continue
            node.add_ssa_ir(Phi(LocalIRVariable(variable), nodes))
        for variable, nodes in node.phi_origins_state_variables.values():
            if len(nodes) < 2:
                continue
            # if not is_used_later(node, variable.name, []):
            #    continue
            node.add_ssa_ir(Phi(StateIRVariable(variable), nodes))

    init_local_variables_instances = {}
    for v in function.parameters:
        if v.name:
            new_var = LocalIRVariable(v)
            function.add_parameter_ssa(new_var)
            if new_var.is_storage:
                fake_variable = LocalIRVariable(v)
                fake_variable.name = "STORAGE_" + fake_variable.name
                fake_variable.set_location("reference_to_storage")
                new_var.refers_to = {fake_variable}
                init_local_variables_instances[fake_variable.name] = fake_variable
            init_local_variables_instances[v.name] = new_var

    for v in function.returns:
        if v.name:
            new_var = LocalIRVariable(v)
            function.add_return_ssa(new_var)
            if new_var.is_storage:
                fake_variable = LocalIRVariable(v)
                fake_variable.name = "STORAGE_" + fake_variable.name
                fake_variable.set_location("reference_to_storage")
                new_var.refers_to = {fake_variable}
                init_local_variables_instances[fake_variable.name] = fake_variable
            init_local_variables_instances[v.name] = new_var

    all_init_local_variables_instances = dict(init_local_variables_instances)

    init_state_variables_instances = dict(all_state_variables_instances)

    initiate_all_local_variables_instances(
        function.nodes,
        init_local_variables_instances,
        all_init_local_variables_instances,
    )

    generate_ssa_irs(
        function.entry_point,
        dict(init_local_variables_instances),
        all_init_local_variables_instances,
        dict(init_state_variables_instances),
        all_state_variables_instances,
        init_local_variables_instances,
        [],
    )

    fix_phi_rvalues_and_storage_ref(
        function.entry_point,
        dict(init_local_variables_instances),
        all_init_local_variables_instances,
        dict(init_state_variables_instances),
        all_state_variables_instances,
        init_local_variables_instances,
    )


def generate_ssa_irs(
    node: Node,
    local_variables_instances: dict[str, LocalIRVariable],
    all_local_variables_instances: dict[str, LocalIRVariable],
    state_variables_instances: dict[str, StateIRVariable],
    all_state_variables_instances: dict[str, StateIRVariable],
    init_local_variables_instances: dict[str, LocalIRVariable],
    visited: list[Node],
) -> None:
    if node in visited:
        return

    if node.type in [NodeType.ENDIF, NodeType.ENDLOOP] and any(
        father not in visited for father in node.fathers
    ):
        return

    # visited is shared
    visited.append(node)

    for ir in node.irs_ssa:
        assert isinstance(ir, Phi)
        update_lvalue(
            ir,
            node,
            local_variables_instances,
            all_local_variables_instances,
            state_variables_instances,
            all_state_variables_instances,
        )

    # these variables are lived only during the liveness of the block
    # They dont need phi function
    temporary_variables_instances = {}
    reference_variables_instances = {}
    tuple_variables_instances = {}

    for ir in node.irs:
        new_ir = copy_ir(
            ir,
            local_variables_instances,
            state_variables_instances,
            temporary_variables_instances,
            reference_variables_instances,
            tuple_variables_instances,
            all_local_variables_instances,
        )

        new_ir.set_expression(ir.expression)
        new_ir.set_node(ir.node)

        update_lvalue(
            new_ir,
            node,
            local_variables_instances,
            all_local_variables_instances,
            state_variables_instances,
            all_state_variables_instances,
        )

        if new_ir:
            node.add_ssa_ir(new_ir)

            if isinstance(ir, (InternalCall, HighLevelCall, InternalDynamicCall, LowLevelCall)):
                if isinstance(ir, LibraryCall):
                    continue
                for variable in all_state_variables_instances.values():
                    if not is_used_later(node, variable):
                        continue
                    new_var = StateIRVariable(variable)
                    new_var.index = all_state_variables_instances[variable.canonical_name].index + 1
                    all_state_variables_instances[variable.canonical_name] = new_var
                    state_variables_instances[variable.canonical_name] = new_var
                    phi_ir = PhiCallback(new_var, {node}, new_ir, variable)
                    # rvalues are fixed in solc_parsing.declaration.function
                    node.add_ssa_ir(phi_ir)

            if isinstance(new_ir, (Assignment, Binary)):
                if isinstance(new_ir.lvalue, LocalIRVariable):
                    if new_ir.lvalue.is_storage:
                        if isinstance(new_ir.rvalue, ReferenceVariable):
                            refers_to = new_ir.rvalue.points_to_origin
                            new_ir.lvalue.add_refers_to(refers_to)
                        # Discard Constant and TopLevelVariable
                        # This can happen on yul, like
                        # assembly { var.slot = some_value }
                        # Here we do not keep track of the references as we do not track
                        # such low level manipulation
                        # However we could extend our storage model to do so in the future
                        elif not isinstance(new_ir.rvalue, (Constant, TopLevelVariable)):
                            new_ir.lvalue.add_refers_to(new_ir.rvalue)

    for succ in node.dominator_successors:
        generate_ssa_irs(
            succ,
            dict(local_variables_instances),
            all_local_variables_instances,
            dict(state_variables_instances),
            all_state_variables_instances,
            init_local_variables_instances,
            visited,
        )

    for dominated in node.dominance_frontier:
        generate_ssa_irs(
            dominated,
            dict(local_variables_instances),
            all_local_variables_instances,
            dict(state_variables_instances),
            all_state_variables_instances,
            init_local_variables_instances,
            visited,
        )


# endregion
###################################################################################
###################################################################################
# region Helpers
###################################################################################
###################################################################################


def last_name(
    n: Node,
    var: StateIRVariable | LocalIRVariable,
    init_vars: dict[str, LocalIRVariable],
) -> StateIRVariable | LocalIRVariable:
    candidates = []
    # Todo optimize by creating a variables_ssa_written attribute
    for ir_ssa in n.irs_ssa:
        if isinstance(ir_ssa, OperationWithLValue):
            lvalue = ir_ssa.lvalue
            while isinstance(lvalue, ReferenceVariable):
                lvalue = lvalue.points_to
            if lvalue and lvalue.name == var.name:
                candidates.append(lvalue)
    if n.variable_declaration and n.variable_declaration.name == var.name:
        candidates.append(LocalIRVariable(n.variable_declaration))
    if n.type == NodeType.ENTRYPOINT:
        if var.name in init_vars:
            candidates.append(init_vars[var.name])
    assert candidates
    return max(candidates, key=lambda v: v.index)


def is_used_later(
    initial_node: Node,
    variable: StateIRVariable | LocalVariable | TemporaryVariableSSA,
) -> bool:
    # TODO: does not handle the case where its read and written in the declaration node
    # It can be problematic if this happens in a loop/if structure
    # Ex:
    # for(;true;){
    #   if(true){
    #     uint a = a;
    #    }
    #     ..
    to_explore = {initial_node}
    explored = set()

    while to_explore:
        node = to_explore.pop()
        explored.add(node)
        if isinstance(variable, LocalVariable):
            if any(v.name == variable.name for v in node.local_variables_read):
                return True
            if any(v.name == variable.name for v in node.local_variables_written):
                return False
        if isinstance(variable, StateVariable):
            if any(
                v.name == variable.name and v.contract == variable.contract
                for v in node.state_variables_read
            ):
                return True
            if any(
                v.name == variable.name and v.contract == variable.contract
                for v in node.state_variables_written
            ):
                return False
        for son in node.sons:
            if son not in explored:
                to_explore.add(son)

    return False


# endregion
###################################################################################
###################################################################################
# region Update operation
###################################################################################
###################################################################################


def update_lvalue(
    new_ir: Operation,
    node: Node,
    local_variables_instances: dict[str, LocalIRVariable],
    all_local_variables_instances: dict[str, LocalIRVariable],
    state_variables_instances: dict[str, StateIRVariable],
    all_state_variables_instances: dict[str, StateIRVariable],
) -> None:
    if isinstance(new_ir, OperationWithLValue):
        lvalue = new_ir.lvalue
        update_through_ref = False
        if isinstance(new_ir, (Assignment, Binary)):
            if isinstance(lvalue, ReferenceVariable):
                update_through_ref = True
                while isinstance(lvalue, ReferenceVariable):
                    lvalue = lvalue.points_to
        if isinstance(lvalue, (LocalIRVariable, StateIRVariable)):
            if isinstance(lvalue, LocalIRVariable):
                new_var = LocalIRVariable(lvalue)
                new_var.index = all_local_variables_instances[lvalue.name].index + 1
                all_local_variables_instances[lvalue.name] = new_var
                local_variables_instances[lvalue.name] = new_var
            else:
                new_var = StateIRVariable(lvalue)
                new_var.index = all_state_variables_instances[lvalue.canonical_name].index + 1
                all_state_variables_instances[lvalue.canonical_name] = new_var
                state_variables_instances[lvalue.canonical_name] = new_var
            if update_through_ref:
                phi_operation = Phi(new_var, {node})
                phi_operation.rvalues = [lvalue]
                node.add_ssa_ir(phi_operation)
            if not isinstance(new_ir.lvalue, ReferenceVariable):
                new_ir.lvalue = new_var
            else:
                to_update = new_ir.lvalue
                while isinstance(to_update.points_to, ReferenceVariable):
                    to_update = to_update.points_to
                to_update.points_to = new_var


# endregion
###################################################################################
###################################################################################
# region Initialization
###################################################################################
###################################################################################


def initiate_all_local_variables_instances(
    nodes: list[Node],
    local_variables_instances: dict[str, LocalIRVariable],
    all_local_variables_instances: dict[str, LocalIRVariable],
) -> None:
    for node in nodes:
        if node.variable_declaration:
            new_var = LocalIRVariable(node.variable_declaration)
            if new_var.name in all_local_variables_instances:
                new_var.index = all_local_variables_instances[new_var.name].index + 1
            local_variables_instances[node.variable_declaration.name] = new_var
            all_local_variables_instances[node.variable_declaration.name] = new_var


# endregion
###################################################################################
###################################################################################
# region Phi Operations
###################################################################################
###################################################################################


def fix_phi_rvalues_and_storage_ref(
    node: Node,
    local_variables_instances: dict[str, LocalIRVariable],
    all_local_variables_instances: dict[str, LocalIRVariable],
    state_variables_instances: dict[str, StateIRVariable],
    all_state_variables_instances: dict[str, StateIRVariable],
    init_local_variables_instances: dict[str, LocalIRVariable],
) -> None:
    for ir in node.irs_ssa:
        if isinstance(ir, (Phi)) and not ir.rvalues:
            variables = [
                last_name(dst, ir.lvalue, init_local_variables_instances) for dst in ir.nodes
            ]
            ir.rvalues = variables
        if isinstance(ir, (Phi, PhiCallback)):
            if isinstance(ir.lvalue, LocalIRVariable):
                if ir.lvalue.is_storage:
                    l = [v.refers_to for v in ir.rvalues]
                    l = [item for sublist in l for item in sublist]
                    ir.lvalue.refers_to = set(l)

        if isinstance(ir, (Assignment, Binary)):
            if isinstance(ir.lvalue, ReferenceVariable):
                origin = ir.lvalue.points_to_origin

                if isinstance(origin, LocalIRVariable):
                    if origin.is_storage:
                        for refers_to in origin.refers_to:
                            phi_ir = Phi(refers_to, {node})
                            phi_ir.rvalues = [origin]
                            node.add_ssa_ir(phi_ir)
                            update_lvalue(
                                phi_ir,
                                node,
                                local_variables_instances,
                                all_local_variables_instances,
                                state_variables_instances,
                                all_state_variables_instances,
                            )
    for succ in node.dominator_successors:
        fix_phi_rvalues_and_storage_ref(
            succ,
            dict(local_variables_instances),
            all_local_variables_instances,
            dict(state_variables_instances),
            all_state_variables_instances,
            init_local_variables_instances,
        )


def add_phi_origins(
    node: Node,
    local_variables_definition: dict[str, tuple[LocalVariable, Node]],
    state_variables_definition: dict[str, tuple[StateVariable, Node]],
) -> None:
    # Add new key to local_variables_definition
    # The key is the variable_name
    # The value is (variable_instance, the node where its written)
    # We keep the instance as we want to avoid to add __hash__ on v.name in Variable
    # That might work for this used, but could create collision for other uses
    local_variables_definition = dict(
        local_variables_definition,
        **{v.name: (v, node) for v in node.local_variables_written},
    )
    state_variables_definition = dict(
        state_variables_definition,
        **{v.canonical_name: (v, node) for v in node.state_variables_written},
    )

    # For unini variable declaration
    if (
        node.variable_declaration
        and node.variable_declaration.name not in local_variables_definition
    ):
        local_variables_definition[node.variable_declaration.name] = (
            node.variable_declaration,
            node,
        )

    # filter length of successors because we have node with one successor
    # while most of the ssa textbook would represent following nodes as one
    if node.dominance_frontier and len(node.dominator_successors) != 1:
        for phi_node in node.dominance_frontier:
            for _, (variable, n) in local_variables_definition.items():
                phi_node.add_phi_origin_local_variable(variable, n)
            for _, (variable, n) in state_variables_definition.items():
                phi_node.add_phi_origin_state_variable(variable, n)

    if not node.dominator_successors:
        return
    for succ in node.dominator_successors:
        add_phi_origins(succ, local_variables_definition, state_variables_definition)


# endregion
###################################################################################
###################################################################################
# region IR copy
###################################################################################
###################################################################################


def get(
    variable,
    local_variables_instances: dict[str, LocalIRVariable],
    state_variables_instances: dict[str, StateIRVariable],
    temporary_variables_instances: dict[int, TemporaryVariableSSA],
    reference_variables_instances: dict[int, ReferenceVariableSSA],
    tuple_variables_instances: dict[int, TupleVariableSSA],
    all_local_variables_instances: dict[str, LocalIRVariable],
):
    # variable can be None
    # for example, on LowLevelCall, ir.lvalue can be none
    if variable is None:
        return None
    if isinstance(variable, LocalVariable):
        if variable.name in local_variables_instances:
            return local_variables_instances[variable.name]
        new_var = LocalIRVariable(variable)
        local_variables_instances[variable.name] = new_var
        all_local_variables_instances[variable.name] = new_var
        return new_var
    if isinstance(variable, StateVariable) and variable.canonical_name in state_variables_instances:
        return state_variables_instances[variable.canonical_name]
    if isinstance(variable, ReferenceVariable):
        if variable.index not in reference_variables_instances:
            new_variable = ReferenceVariableSSA(variable)
            if variable.points_to:
                new_variable.points_to = get(
                    variable.points_to,
                    local_variables_instances,
                    state_variables_instances,
                    temporary_variables_instances,
                    reference_variables_instances,
                    tuple_variables_instances,
                    all_local_variables_instances,
                )
            new_variable.set_type(variable.type)
            reference_variables_instances[variable.index] = new_variable
        return reference_variables_instances[variable.index]
    if isinstance(variable, TemporaryVariable):
        if variable.index not in temporary_variables_instances:
            new_variable = TemporaryVariableSSA(variable)
            new_variable.set_type(variable.type)
            temporary_variables_instances[variable.index] = new_variable
        return temporary_variables_instances[variable.index]
    if isinstance(variable, TupleVariable):
        if variable.index not in tuple_variables_instances:
            new_variable = TupleVariableSSA(variable)
            new_variable.set_type(variable.type)
            tuple_variables_instances[variable.index] = new_variable
        return tuple_variables_instances[variable.index]
    assert isinstance(
        variable,
        (
            Constant,
            SolidityVariable,
            Contract,
            Enum,
            SolidityFunction,
            Structure,
            Function,
            Type,
            SolidityImportPlaceHolder,
            TopLevelVariable,
        ),
    )  # type for abi.decode(.., t)
    return variable


def get_variable(ir: Operation, f: Callable, *instances):
    variable = f(ir)
    variable = get(variable, *instances)
    return variable


def _get_traversal(values: list[Any], *instances) -> list[Any]:
    ret = []

    for v in values:
        if isinstance(v, list):
            v = _get_traversal(v, *instances)
        else:
            v = get(v, *instances)
        ret.append(v)
    return ret


def get_arguments(ir: Call, *instances) -> list[Any]:
    return _get_traversal(ir.arguments, *instances)


def get_rec_values(
    ir: InitArray | Return | NewArray,
    f: Callable,
    *instances,
) -> list[Variable]:
    # Use by InitArray and NewArray
    # Potential recursive array(s)
    ori_init_values = f(ir)

    return _get_traversal(ori_init_values, *instances)


def copy_ir(ir: Operation, *instances) -> Operation:
    """
    Args:
        ir (Operation)
        local_variables_instances(dict(str -> LocalVariable))
        state_variables_instances(dict(str -> StateVariable))
        temporary_variables_instances(dict(int -> Variable))
        reference_variables_instances(dict(int -> Variable))

    Note: temporary and reference can be indexed by int, as they dont need phi functions
    """
    if isinstance(ir, Assignment):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        rvalue = get_variable(ir, lambda x: x.rvalue, *instances)
        variable_return_type = ir.variable_return_type
        return Assignment(lvalue, rvalue, variable_return_type)
    if isinstance(ir, Binary):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        variable_left = get_variable(ir, lambda x: x.variable_left, *instances)
        variable_right = get_variable(ir, lambda x: x.variable_right, *instances)
        operation_type = ir.type
        return Binary(lvalue, variable_left, variable_right, operation_type)
    if isinstance(ir, CodeSize):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        value = get_variable(ir, lambda x: x.value, *instances)
        return CodeSize(value, lvalue)
    if isinstance(ir, Condition):
        val = get_variable(ir, lambda x: x.value, *instances)
        return Condition(val)
    if isinstance(ir, Delete):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        variable = get_variable(ir, lambda x: x.variable, *instances)
        return Delete(lvalue, variable)
    if isinstance(ir, EventCall):
        name = ir.name
        new_ir = EventCall(name)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, HighLevelCall):  # include LibraryCall
        destination = get_variable(ir, lambda x: x.destination, *instances)
        function_name = ir.function_name
        nbr_arguments = ir.nbr_arguments
        names = ir.names
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        type_call = ir.type_call
        if isinstance(ir, LibraryCall):
            new_ir = LibraryCall(
                destination, function_name, nbr_arguments, lvalue, type_call, names=names
            )
        else:
            new_ir = HighLevelCall(
                destination, function_name, nbr_arguments, lvalue, type_call, names=names
            )
        new_ir.call_id = ir.call_id
        new_ir.call_value = get_variable(ir, lambda x: x.call_value, *instances)
        new_ir.call_gas = get_variable(ir, lambda x: x.call_gas, *instances)
        new_ir.arguments = get_arguments(ir, *instances)
        new_ir.function = ir.function
        return new_ir
    if isinstance(ir, Index):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        variable_left = get_variable(ir, lambda x: x.variable_left, *instances)
        variable_right = get_variable(ir, lambda x: x.variable_right, *instances)
        return Index(lvalue, variable_left, variable_right)
    if isinstance(ir, InitArray):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        init_values = get_rec_values(ir, lambda x: x.init_values, *instances)
        return InitArray(init_values, lvalue)
    if isinstance(ir, InternalCall):
        function = ir.function
        nbr_arguments = ir.nbr_arguments
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        type_call = ir.type_call
        names = ir.names
        new_ir = InternalCall(function, nbr_arguments, lvalue, type_call, names=names)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, InternalDynamicCall):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        function = get_variable(ir, lambda x: x.function, *instances)
        function_type = ir.function_type
        new_ir = InternalDynamicCall(lvalue, function, function_type)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, LowLevelCall):
        destination = get_variable(ir, lambda x: x.destination, *instances)
        function_name = ir.function_name
        nbr_arguments = ir.nbr_arguments
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        type_call = ir.type_call
        new_ir = LowLevelCall(destination, function_name, nbr_arguments, lvalue, type_call)
        new_ir.call_id = ir.call_id
        new_ir.call_value = get_variable(ir, lambda x: x.call_value, *instances)
        new_ir.call_gas = get_variable(ir, lambda x: x.call_gas, *instances)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, Member):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        variable_left = get_variable(ir, lambda x: x.variable_left, *instances)
        variable_right = get_variable(ir, lambda x: x.variable_right, *instances)
        return Member(variable_left, variable_right, lvalue)
    if isinstance(ir, NewArray):
        array_type = ir.array_type
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        new_ir = NewArray(array_type, lvalue)
        new_ir.arguments = get_rec_values(ir, lambda x: x.arguments, *instances)
        return new_ir
    if isinstance(ir, NewElementaryType):
        new_type = ir.type
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        new_ir = NewElementaryType(new_type, lvalue)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, NewContract):
        contract_name = ir.contract_name
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        new_ir = NewContract(contract_name, lvalue)
        new_ir.arguments = get_arguments(ir, *instances)
        new_ir.call_value = get_variable(ir, lambda x: x.call_value, *instances)
        new_ir.call_salt = get_variable(ir, lambda x: x.call_salt, *instances)
        return new_ir
    if isinstance(ir, NewStructure):
        structure = ir.structure
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        names = ir.names
        new_ir = NewStructure(structure, lvalue, names=names)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, Nop):
        return Nop()
    if isinstance(ir, Return):
        values = get_rec_values(ir, lambda x: x.values, *instances)
        return Return(values)
    if isinstance(ir, Send):
        destination = get_variable(ir, lambda x: x.destination, *instances)
        value = get_variable(ir, lambda x: x.call_value, *instances)
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        return Send(destination, value, lvalue)
    if isinstance(ir, SolidityCall):
        function = ir.function
        nbr_arguments = ir.nbr_arguments
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        type_call = ir.type_call
        new_ir = SolidityCall(function, nbr_arguments, lvalue, type_call)
        new_ir.arguments = get_arguments(ir, *instances)
        return new_ir
    if isinstance(ir, Transfer):
        destination = get_variable(ir, lambda x: x.destination, *instances)
        value = get_variable(ir, lambda x: x.call_value, *instances)
        return Transfer(destination, value)
    if isinstance(ir, TypeConversion):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        variable = get_variable(ir, lambda x: x.variable, *instances)
        variable_type = ir.type
        return TypeConversion(lvalue, variable, variable_type)
    if isinstance(ir, Unary):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        rvalue = get_variable(ir, lambda x: x.rvalue, *instances)
        operation_type = ir.type
        return Unary(lvalue, rvalue, operation_type)
    if isinstance(ir, Unpack):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        tuple_var = get_variable(ir, lambda x: x.tuple, *instances)
        idx = ir.index
        return Unpack(lvalue, tuple_var, idx)
    if isinstance(ir, Length):
        lvalue = get_variable(ir, lambda x: x.lvalue, *instances)
        value = get_variable(ir, lambda x: x.value, *instances)
        return Length(value, lvalue)

    raise SlithIRError(f"Impossible ir copy on {ir} ({type(ir)})")


# endregion



import argparse
import logging
from crytic_compile import cryticparser
from slither import Slither

logging.basicConfig()
logging.getLogger("Slither").setLevel(logging.INFO)

logger = logging.getLogger("Slither-demo")


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(description="Demo", usage="slither-demo filename")

    parser.add_argument(
        "filename", help="The filename of the contract or truffle directory to analyze."
    )

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    # Perform slither analysis on the given filename
    _slither = Slither(args.filename, **vars(args))

    logger.info("Analysis done!")


if __name__ == "__main__":
    main()



from slither.core.cfg.node import Node, NodeType
from slither.core.solidity_types import ElementaryType
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import (
    OperationWithLValue,
    HighLevelCall,
    InternalDynamicCall,
    InternalCall,
    LowLevelCall,
    Operation,
)
from slither.slithir.variables import ReferenceVariable, TemporaryVariable, TupleVariable
from slither.slithir.variables.variable import SlithIRVariable
from slither.utils.output import Output


def _remove_states(written: dict[Variable, Node]) -> None:
    for key in list(written.keys()):
        if isinstance(key, StateVariable):
            del written[key]


def _handle_ir(
    ir: Operation,
    written: dict[Variable, Node],
    ret: list[tuple[Variable, Node, Node]],
) -> None:
    if isinstance(ir, (HighLevelCall, InternalDynamicCall, LowLevelCall)):
        _remove_states(written)

    if isinstance(ir, InternalCall):
        if not ir.function:
            return
        if ir.function.all_high_level_calls() or ir.function.all_library_calls():
            _remove_states(written)

        all_read = ir.function.all_state_variables_read()
        for read in all_read:
            if (
                isinstance(read, Variable)
                and isinstance(read.type, ElementaryType)
                and not isinstance(read, SlithIRVariable)
                and read in written
            ):
                del written[read]

    for read in ir.read:
        if (
            isinstance(read, Variable)
            and isinstance(read.type, ElementaryType)
            and not isinstance(read, SlithIRVariable)
            and read in written
        ):
            del written[read]

    if isinstance(ir, OperationWithLValue):
        # Until we have a better handling of mapping/array we only look for simple types
        if (
            ir.lvalue
            and isinstance(ir.lvalue.type, ElementaryType)
            and not isinstance(ir.lvalue, (ReferenceVariable, TemporaryVariable, TupleVariable))
        ):
            if ir.lvalue.name == "_":
                return
            if ir.lvalue in written:
                ret.append((ir.lvalue, written[ir.lvalue], ir.node))
            written[ir.lvalue] = ir.node


def _detect_write_after_write(
    node: Node,
    explored: set[Node],
    written: dict[Variable, Node],
    ret: list[tuple[Variable, Node, Node]],
) -> None:
    if node in explored:
        return

    explored.add(node)

    # We could report write after write for, but this lead to a lot of FP due to the initialization to zero pattern:
    # uint a = 0;
    # a = 10;
    # To do better, we could filter out if the variable is init to zero
    if node.type != NodeType.VARIABLE:
        for ir in node.irs:
            _handle_ir(ir, written, ret)

    if len(node.sons) > 1:
        written = {}
    for son in node.sons:
        _detect_write_after_write(son, explored, dict(written), ret)


class WriteAfterWrite(AbstractDetector):
    ARGUMENT = "write-after-write"
    HELP = "Unused write"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#write-after-write"

    WIKI_TITLE = "Write after write"
    WIKI_DESCRIPTION = """Detects variables that are written but never read and written again."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
    ```solidity
    contract Buggy{
        function my_func() external initializer{
            // ...
            a = b;
            a = c;
            // ..
        }
    }
    ```
    `a` is first assigned to `b`, and then to `c`. As a result the first write does nothing."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Fix or remove the writes."""

    def _detect(self) -> list[Output]:
        results = []

        for contract in self.compilation_unit.contracts_derived:
            for function in contract.functions:
                if function.entry_point:
                    ret: list[tuple[Variable, Node, Node]] = []
                    _detect_write_after_write(function.entry_point, set(), {}, ret)
                    for var, node1, node2 in ret:
                        info: DETECTOR_INFO = [
                            var,
                            " is written in both\n\t",
                            node1,
                            "\n\t",
                            node2,
                            "\n",
                        ]

                        res = self.generate_result(info)
                        results.append(res)

        return results



import sys

try:
    import numpy as np
except ImportError:
    print("ERROR: in order to use slither-simil, you need to install numpy:")
    print("$ pip3 install numpy --user\n")
    sys.exit(-1)


def similarity(v1, v2):
    n1 = np.linalg.norm(v1)
    n2 = np.linalg.norm(v2)
    return np.dot(v1, v2) / n1 / n2



from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.utils.myprettytable import MyPrettyTable


def output_mutators(mutators_classes: list[type[AbstractMutator]]) -> None:
    mutators_list = []
    for detector in mutators_classes:
        argument = detector.NAME
        help_info = detector.HELP
        mutators_list.append((argument, help_info))
    table = MyPrettyTable(["Num", "Name", "What it Does"])

    # Sort by class
    mutators_list = sorted(mutators_list, key=lambda element: (element[0]))
    idx = 1
    for argument, help_info in mutators_list:
        table.add_row([str(idx), argument, help_info])
        idx = idx + 1
    print(table)



from pathlib import Path
from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"

ALL_TESTS = list(Path(TEST_DATA_DIR).glob("*.vy"))


def pytest_generate_tests(metafunc):
    test_cases = []
    for test_file in ALL_TESTS:
        sl = Slither(test_file.as_posix())
        for contract in sl.contracts:
            if contract.is_interface:
                continue
            for func_or_modifier in contract.functions:
                test_cases.append(
                    (func_or_modifier.canonical_name, func_or_modifier.slithir_cfg_to_dot_str())
                )

    metafunc.parametrize("test_case", test_cases, ids=lambda x: x[0])


def test_vyper_cfgir(test_case, snapshot):
    assert snapshot() == test_case[1]



"""
Module printing evm mapping of the contract
"""

import logging
from typing import Union

from slither.printers.abstract_printer import AbstractPrinter
from slither.core.declarations.function import Function
from slither.core.declarations.modifier import Modifier
from slither.analyses.evm import (
    generate_source_to_evm_ins_mapping,
    load_evm_cfg_builder,
)
from slither.utils.colors import blue, green, magenta, red


logger: logging.Logger = logging.getLogger("EVMPrinter")


def _extract_evm_info(slither):
    """
    Extract evm information for all derived contracts using evm_cfg_builder

    Returns: evm CFG and Solidity source to Program Counter (pc) mapping
    """

    evm_info = {}

    CFG = load_evm_cfg_builder()

    for contract in slither.contracts_derived:
        contract_bytecode_runtime = contract.file_scope.bytecode_runtime(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )

        if not contract_bytecode_runtime:
            logger.info(
                "Contract %s (abstract: %r) has no bytecode runtime, skipping. ",
                contract.name,
                contract.is_abstract,
            )
            evm_info["empty", contract.name] = True
            continue

        contract_srcmap_runtime = contract.file_scope.srcmap_runtime(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        cfg = CFG(contract_bytecode_runtime)
        evm_info["cfg", contract.name] = cfg
        evm_info["mapping", contract.name] = generate_source_to_evm_ins_mapping(
            cfg.instructions,
            contract_srcmap_runtime,
            slither,
            contract.source_mapping.filename.absolute,
        )

        contract_bytecode_init = contract.file_scope.bytecode_init(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        contract_srcmap_init = contract.file_scope.srcmap_init(
            contract.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        cfg_init = CFG(contract_bytecode_init)

        evm_info["cfg_init", contract.name] = cfg_init
        evm_info["mapping_init", contract.name] = generate_source_to_evm_ins_mapping(
            cfg_init.instructions,
            contract_srcmap_init,
            slither,
            contract.source_mapping.filename.absolute,
        )

    return evm_info


class PrinterEVM(AbstractPrinter):
    ARGUMENT = "evm"
    HELP = "Print the evm instructions of nodes in functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#evm"

    def build_element_node_str(
        self,
        element: Union["Modifier", "Function"],
        contract_pcs: dict[int, list[int]],
        contract_cfg,
    ) -> str:
        element_file = self.slither.source_code[
            element.contract_declarer.source_mapping.filename.absolute
        ].splitlines()

        return_string = ""
        for node in element.nodes:
            return_string += green(f"\t\tNode: {node}\n")
            node_source_line = node.source_mapping.lines[0]
            return_string += green(
                f"\t\tSource line {node_source_line}: {element_file[node_source_line - 1].rstrip()}\n"
            )

            return_string += magenta("\t\tEVM Instructions:\n")
            node_pcs = contract_pcs.get(node_source_line, [])
            for pc in node_pcs:
                return_string += magenta(
                    f"\t\t\t{hex(pc)}: {contract_cfg.get_instruction_at(pc)}\n"
                )

        return return_string

    def output(self, _filename):
        """
        _filename is not used
        Args:
            _filename(string)
        """

        txt = ""
        if not self.slither.crytic_compile:
            txt = "The EVM printer requires to compile with crytic-compile"
            self.info(red(txt))
            res = self.generate_output(txt)
            return res
        evm_info = _extract_evm_info(self.slither)

        for contract in self.slither.contracts_derived:
            txt += blue(f"Contract {contract.name}\n")

            if evm_info.get(("empty", contract.name), False):
                txt += "\tempty contract\n"
                continue

            for function in contract.functions:
                txt += blue(f"\tFunction {function.canonical_name}\n")

                txt += self.build_element_node_str(
                    function,
                    evm_info["mapping", contract.name]
                    if not function.is_constructor
                    else evm_info["mapping_init", contract.name],
                    evm_info["cfg", contract.name]
                    if not function.is_constructor
                    else evm_info["cfg_init", contract.name],
                )

            for modifier in contract.modifiers:
                txt += blue(f"\tModifier {modifier.canonical_name}\n")

                txt += self.build_element_node_str(
                    modifier,
                    evm_info["mapping", contract.name],
                    evm_info["cfg", contract.name],
                )

        self.info(txt)
        res = self.generate_output(txt)
        return res



import logging
import os

from slither import Slither
from slither.core.declarations import (
    Structure,
    Enum,
    SolidityVariableComposed,
    SolidityVariable,
    Function,
)
from slither.core.solidity_types import (
    ElementaryType,
    ArrayType,
    MappingType,
    UserDefinedType,
)
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.local_variable_init_from_tuple import LocalVariableInitFromTuple
from slither.core.variables.state_variable import StateVariable
from slither.slithir.operations import (
    Assignment,
    Index,
    Member,
    Length,
    Binary,
    Unary,
    Condition,
    NewArray,
    NewStructure,
    NewContract,
    NewElementaryType,
    SolidityCall,
    Delete,
    EventCall,
    LibraryCall,
    InternalDynamicCall,
    HighLevelCall,
    LowLevelCall,
    TypeConversion,
    Return,
    Transfer,
    Send,
    Unpack,
    InitArray,
    InternalCall,
)
from slither.slithir.variables import (
    TemporaryVariable,
    TupleVariable,
    Constant,
    ReferenceVariable,
)
from .cache import load_cache

simil_logger = logging.getLogger("Slither-simil")
compiler_logger = logging.getLogger("CryticCompile")
compiler_logger.setLevel(logging.CRITICAL)
slither_logger = logging.getLogger("Slither")
slither_logger.setLevel(logging.CRITICAL)


def parse_target(target: str | None) -> tuple[str | None, str | None]:
    if target is None:
        return None, None

    parts = target.split(".")
    if len(parts) == 1:
        return None, parts[0]
    if len(parts) == 2:
        return parts[0], parts[1]
    simil_logger.error("Invalid target. It should be 'function' or 'Contract.function'")
    return None, None


def load_and_encode(infile: str, vmodel, ext=None, nsamples=None, **kwargs):
    r = {}
    if infile.endswith(".npz"):
        r = load_cache(infile, nsamples=nsamples)
    else:
        contracts = load_contracts(infile, ext=ext, nsamples=nsamples)
        for contract in contracts:
            for x, ir in encode_contract(contract, **kwargs).items():
                if ir != []:
                    y = " ".join(ir)
                    r[x] = vmodel.get_sentence_vector(y)

    return r


def load_contracts(dirname: str, ext: str | None = None, nsamples: int | None = None) -> list[str]:
    r = []
    walk = list(os.walk(dirname))
    for x, y, files in walk:
        for f in files:
            if ext is None or f.endswith(ext):
                r.append(x + "/".join(y) + "/" + f)

    if nsamples is None:
        return r

    # TODO: shuffle
    return r[:nsamples]


def ntype(_type):
    if isinstance(_type, ElementaryType):
        _type = str(_type)
    elif isinstance(_type, ArrayType):
        if isinstance(_type.type, ElementaryType):
            _type = str(_type)
        else:
            _type = "user_defined_array"
    elif isinstance(_type, Structure):
        _type = str(_type)
    elif isinstance(_type, Enum):
        _type = str(_type)
    elif isinstance(_type, MappingType):
        _type = str(_type)
    elif isinstance(_type, UserDefinedType):
        _type = "user_defined_type"  # TODO: this could be Contract, Enum or Struct
    else:
        _type = str(_type)

    _type = _type.replace(" memory", "")
    _type = _type.replace(" storage ref", "")

    if "struct" in _type:
        return "struct"
    if "enum" in _type:
        return "enum"
    if "tuple" in _type:
        return "tuple"
    if "contract" in _type:
        return "contract"
    if "mapping" in _type:
        return "mapping"
    return _type.replace(" ", "_")


def encode_ir(ir):
    # operations
    if isinstance(ir, Assignment):
        return f"({encode_ir(ir.lvalue)}):=({encode_ir(ir.rvalue)})"
    if isinstance(ir, Index):
        return f"index({ntype(ir.index_type)})"
    if isinstance(ir, Member):
        return "member"  # .format(ntype(ir._type))
    if isinstance(ir, Length):
        return "length"
    if isinstance(ir, Binary):
        return f"binary({ir.type!s})"
    if isinstance(ir, Unary):
        return f"unary({ir.type!s})"
    if isinstance(ir, Condition):
        return f"condition({encode_ir(ir.value)})"
    if isinstance(ir, NewStructure):
        return "new_structure"
    if isinstance(ir, NewContract):
        return "new_contract"
    if isinstance(ir, NewArray):
        return f"new_array({ntype(ir.array_type)})"
    if isinstance(ir, NewElementaryType):
        return f"new_elementary({ntype(ir.type)})"
    if isinstance(ir, Delete):
        return f"delete({encode_ir(ir.lvalue)},{encode_ir(ir.variable)})"
    if isinstance(ir, SolidityCall):
        return f"solidity_call({ir.function.full_name})"
    if isinstance(ir, InternalCall):
        return f"internal_call({ntype(ir.type_call)})"
    if isinstance(ir, EventCall):  # is this useful?
        return "event"
    if isinstance(ir, LibraryCall):
        return "library_call"
    if isinstance(ir, InternalDynamicCall):
        return "internal_dynamic_call"
    if isinstance(ir, HighLevelCall):  # TODO: improve
        return "high_level_call"
    if isinstance(ir, LowLevelCall):  # TODO: improve
        return "low_level_call"
    if isinstance(ir, TypeConversion):
        return f"type_conversion({ntype(ir.type)})"
    if isinstance(ir, Return):  # this can be improved using values
        return "return"  # .format(ntype(ir.type))
    if isinstance(ir, Transfer):
        return f"transfer({encode_ir(ir.call_value)})"
    if isinstance(ir, Send):
        return f"send({encode_ir(ir.call_value)})"
    if isinstance(ir, Unpack):  # TODO: improve
        return "unpack"
    if isinstance(ir, InitArray):  # TODO: improve
        return "init_array"
    if isinstance(ir, Function):  # TODO: investigate this
        return "function_solc"

    # variables
    if isinstance(ir, Constant):
        return f"constant({ntype(ir.type)})"
    if isinstance(ir, SolidityVariableComposed):
        return f"solidity_variable_composed({ir.name})"
    if isinstance(ir, SolidityVariable):
        return f"solidity_variable{ir.name}"
    if isinstance(ir, TemporaryVariable):
        return "temporary_variable"
    if isinstance(ir, ReferenceVariable):
        return f"reference({ntype(ir.type)})"
    if isinstance(ir, LocalVariable):
        return f"local_solc_variable({ir.location})"
    if isinstance(ir, StateVariable):
        return f"state_solc_variable({ntype(ir.type)})"
    if isinstance(ir, LocalVariableInitFromTuple):
        return "local_variable_init_tuple"
    if isinstance(ir, TupleVariable):
        return "tuple_variable"

    # default
    simil_logger.error(type(ir), "is missing encoding!")
    return ""


def encode_contract(cfilename, **kwargs):
    r = {}

    # Init slither
    try:
        slither = Slither(cfilename, **kwargs)
    except Exception:
        simil_logger.error("Compilation failed for %s using %s", cfilename, kwargs["solc"])
        return r

    # Iterate over all the contracts
    for contract in slither.contracts:
        # Iterate over all the functions
        for function in contract.functions_declared:
            if function.nodes == [] or function.is_constructor_variables:
                continue

            x = (cfilename, contract.name, function.name)

            r[x] = []

            # Iterate over the nodes of the function
            for node in function.nodes:
                # Print the Solidity expression of the nodes
                # And the SlithIR operations
                if node.expression:
                    for ir in node.irs:
                        r[x].append(encode_ir(ir))
    return r



"""
Detect incorrect erc721 interface.
"""

from typing import Any
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.utils.output import Output


class IncorrectERC721InterfaceDetection(AbstractDetector):
    """
    Incorrect ERC721 Interface
    """

    ARGUMENT = "erc721-interface"
    HELP = "Incorrect ERC721 interfaces"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#incorrect-erc721-interface"
    )

    WIKI_TITLE = "Incorrect erc721 interface"
    WIKI_DESCRIPTION = "Incorrect return values for `ERC721` functions. A contract compiled with solidity > 0.4.22 interacting with these functions will fail to execute them, as the return value is missing."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Token{
    function ownerOf(uint256 _tokenId) external view returns (bool);
    //...
}
```
`Token.ownerOf` does not return an address like `ERC721` expects. Bob deploys the token. Alice creates a contract that interacts with it but assumes a correct `ERC721` interface implementation. Alice's contract is unable to interact with Bob's contract."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Set the appropriate return values and vtypes for the defined `ERC721` functions."
    )

    @staticmethod
    def incorrect_erc721_interface(
        signature: tuple[str, list[str], list[str]] | tuple[str, list[str], list[Any]],
    ) -> bool:
        (name, parameters, returnVars) = signature

        # ERC721
        if name == "balanceOf" and parameters == ["address"] and returnVars != ["uint256"]:
            return True
        if name == "ownerOf" and parameters == ["uint256"] and returnVars != ["address"]:
            return True
        if (
            name == "safeTransferFrom"
            and parameters == ["address", "address", "uint256", "bytes"]
            and returnVars != []
        ):
            return True
        if (
            name == "safeTransferFrom"
            and parameters == ["address", "address", "uint256"]
            and returnVars != []
        ):
            return True
        if (
            name == "transferFrom"
            and parameters == ["address", "address", "uint256"]
            and returnVars != []
        ):
            return True
        if name == "approve" and parameters == ["address", "uint256"] and returnVars != []:
            return True
        if name == "setApprovalForAll" and parameters == ["address", "bool"] and returnVars != []:
            return True
        if name == "getApproved" and parameters == ["uint256"] and returnVars != ["address"]:
            return True
        if (
            name == "isApprovedForAll"
            and parameters == ["address", "address"]
            and returnVars != ["bool"]
        ):
            return True

        # ERC165 (dependency)
        if name == "supportsInterface" and parameters == ["bytes4"] and returnVars != ["bool"]:
            return True

        return False

    @staticmethod
    def detect_incorrect_erc721_interface(
        contract: Contract,
    ) -> list[FunctionContract | Any]:
        """Detect incorrect ERC721 interface

        Returns:
            list(str) : list of incorrect function signatures
        """

        # Verify this is an ERC721 contract.
        if not contract.is_possible_erc721() or not contract.is_possible_erc20():
            return []

        funcs = contract.functions
        functions = [
            f
            for f in funcs
            if IncorrectERC721InterfaceDetection.incorrect_erc721_interface(f.signature)
        ]
        return functions

    def _detect(self) -> list[Output]:
        """Detect incorrect erc721 interface

        Returns:
            dict: [contract name] = set(str)  events
        """
        results = []
        for c in self.compilation_unit.contracts_derived:
            functions = IncorrectERC721InterfaceDetection.detect_incorrect_erc721_interface(c)
            if functions:
                for function in functions:
                    info: DETECTOR_INFO = [
                        c,
                        " has incorrect ERC721 function interface:",
                        function,
                        "\n",
                    ]
                    res = self.generate_result(info)

                    results.append(res)

        return results



import logging

from crytic_compile import CryticCompile, InvalidCompilation


from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.slither_core import SlitherCore
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.exceptions import SlitherError
from slither.printers.abstract_printer import AbstractPrinter
from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc
from slither.vyper_parsing.vyper_compilation_unit import VyperCompilationUnit
from slither.utils.output import Output
from slither.vyper_parsing.ast.ast import parse

logger = logging.getLogger("Slither")
logging.basicConfig()

logger_detector = logging.getLogger("Detectors")
logger_printer = logging.getLogger("Printers")


def _check_common_things(
    thing_name: str, cls: type, base_cls: type, instances_list: list[type[AbstractDetector]]
) -> None:
    if not issubclass(cls, base_cls) or cls is base_cls:
        raise SlitherError(
            f"You can't register {cls!r} as a {thing_name}. You need to pass a class that inherits from {base_cls.__name__}"
        )

    if any(type(obj) is cls for obj in instances_list):
        raise SlitherError(f"You can't register {cls!r} twice.")


def _update_file_scopes(
    sol_parser: SlitherCompilationUnitSolc,
):
    """
    Since all definitions in a file are exported by default, including definitions from its (transitive) dependencies,
    we can identify all top level items that could possibly be referenced within the file from its exportedSymbols.
    It is not as straightforward for user defined types and functions as well as aliasing. See add_accessible_scopes for more details.
    """
    candidates = sol_parser.compilation_unit.scopes.values()
    learned_something = False
    # Because solc's import allows cycle in the import graph, iterate until we aren't adding new information to the scope.
    while True:
        for candidate in candidates:
            learned_something |= candidate.add_accessible_scopes()
        if not learned_something:
            break
        learned_something = False

    for scope in candidates:
        for refId in scope.exported_symbols:
            if refId in sol_parser.contracts_by_id:
                contract = sol_parser.contracts_by_id[refId]
                scope.contracts[contract.name] = contract
            elif refId in sol_parser.functions_by_id:
                functions = sol_parser.functions_by_id[refId]
                assert len(functions) == 1
                scope.functions.add(functions[0])
            elif refId in sol_parser.imports_by_id:
                import_directive = sol_parser.imports_by_id[refId]
                scope.imports.add(import_directive)
            elif refId in sol_parser.top_level_variables_by_id:
                top_level_variable = sol_parser.top_level_variables_by_id[refId]
                scope.variables[top_level_variable.name] = top_level_variable
            elif refId in sol_parser.top_level_events_by_id:
                top_level_event = sol_parser.top_level_events_by_id[refId]
                scope.events.add(top_level_event)
            elif refId in sol_parser.top_level_structures_by_id:
                top_level_struct = sol_parser.top_level_structures_by_id[refId]
                scope.structures[top_level_struct.name] = top_level_struct
            elif refId in sol_parser.top_level_type_aliases_by_id:
                top_level_type_alias = sol_parser.top_level_type_aliases_by_id[refId]
                scope.type_aliases[top_level_type_alias.name] = top_level_type_alias
            elif refId in sol_parser.top_level_enums_by_id:
                top_level_enum = sol_parser.top_level_enums_by_id[refId]
                scope.enums[top_level_enum.name] = top_level_enum
            elif refId in sol_parser.top_level_errors_by_id:
                top_level_custom_error = sol_parser.top_level_errors_by_id[refId]
                scope.custom_errors.add(top_level_custom_error)
            else:
                logger.error(
                    f"Failed to resolved name for reference id {refId} in {scope.filename.absolute}."
                )


class Slither(SlitherCore):
    def __init__(self, target: str | CryticCompile, **kwargs) -> None:
        """
        Args:
            target (str | CryticCompile)
        Keyword Args:
            solc (str): solc binary location (default 'solc')
            disable_solc_warnings (bool): True to disable solc warnings (default false)
            solc_args (str): solc arguments (default '')
            ast_format (str): ast format (default '--ast-compact-json')
            filter_paths (list(str)): list of path to filter (default [])
            triage_mode (bool): if true, switch to triage mode (default false)
            exclude_dependencies (bool): if true, exclude results that are only related to dependencies
            generate_patches (bool): if true, patches are generated (json output only)
            change_line_prefix (str): Change the line prefix (default #)
                for the displayed source codes (i.e. file.sol#1).

        """
        super().__init__()

        self._disallow_partial: bool = kwargs.get("disallow_partial", False)
        self._skip_assembly: bool = kwargs.get("skip_assembly", False)
        self._show_ignored_findings: bool = kwargs.get("show_ignored_findings", False)

        self.line_prefix = kwargs.get("change_line_prefix", "#")

        # Indicate if Codex related features should be used
        self.codex_enabled = kwargs.get("codex", False)
        self.codex_contracts = kwargs.get("codex_contracts", "all")
        self.codex_model = kwargs.get("codex_model", "text-davinci-003")
        self.codex_temperature = kwargs.get("codex_temperature", 0)
        self.codex_max_tokens = kwargs.get("codex_max_tokens", 300)
        self.codex_log = kwargs.get("codex_log", False)
        self.codex_organization: str | None = kwargs.get("codex_organization")

        self.no_fail = kwargs.get("no_fail", False)

        self._parsers: list[SlitherCompilationUnitSolc] = []
        try:
            if isinstance(target, CryticCompile):
                crytic_compile = target
            else:
                crytic_compile = CryticCompile(target, **kwargs)
            self._crytic_compile = crytic_compile
        except InvalidCompilation as e:
            raise SlitherError(f"Invalid compilation: \n{e!s}")
        for compilation_unit in crytic_compile.compilation_units.values():
            compilation_unit_slither = SlitherCompilationUnit(self, compilation_unit)
            self._compilation_units.append(compilation_unit_slither)

            if compilation_unit_slither.is_vyper:
                vyper_parser = VyperCompilationUnit(compilation_unit_slither)
                for path, ast in compilation_unit.asts.items():
                    ast_nodes = parse(ast["ast"])
                    vyper_parser.parse_module(ast_nodes, path)
                self._parsers.append(vyper_parser)
            else:
                # Solidity specific
                assert compilation_unit_slither.is_solidity
                sol_parser = SlitherCompilationUnitSolc(compilation_unit_slither)
                self._parsers.append(sol_parser)
                for path, ast in compilation_unit.asts.items():
                    sol_parser.parse_top_level_items(ast, path)
                    self.add_source_code(path)

                for contract in sol_parser._underlying_contract_to_parser:
                    if contract.name.startswith("SlitherInternalTopLevelContract"):
                        raise SlitherError(
                            # region multi-line-string
                            """Your codebase has a contract named 'SlitherInternalTopLevelContract'.
        Please rename it, this name is reserved for Slither's internals"""
                            # endregion multi-line
                        )
                    sol_parser._contracts_by_id[contract.id] = contract
                    sol_parser._compilation_unit.contracts.append(contract)

                _update_file_scopes(sol_parser)

        if kwargs.get("generate_patches", False):
            self.generate_patches = True

        self._markdown_root = kwargs.get("markdown_root", "")

        self._detectors = []
        self._printers = []

        filter_paths = kwargs.get("filter_paths", [])
        for p in filter_paths:
            self.add_path_to_filter(p)

        include_paths = kwargs.get("include_paths", [])
        for p in include_paths:
            self.add_path_to_include(p)

        self._exclude_dependencies = kwargs.get("exclude_dependencies", False)

        triage_mode = kwargs.get("triage_mode", False)
        triage_database = kwargs.get("triage_database", "slither.db.json")
        self._triage_mode = triage_mode
        self._previous_results_filename = triage_database

        printers_to_run = kwargs.get("printers_to_run", "")
        if printers_to_run == "echidna":
            self.skip_data_dependency = True

        # Used in inheritance-graph printer
        self.include_interfaces = kwargs.get("include_interfaces", False)

        self._init_parsing_and_analyses(kwargs.get("skip_analyze", False))

    def _init_parsing_and_analyses(self, skip_analyze: bool) -> None:
        for parser in self._parsers:
            try:
                parser.parse_contracts()
            except Exception as e:
                if self.no_fail:
                    continue
                raise e

        # skip_analyze is only used for testing
        if not skip_analyze:
            for parser in self._parsers:
                try:
                    parser.analyze_contracts()
                except Exception as e:
                    if self.no_fail:
                        continue
                    raise e

    @property
    def detectors(self):
        return self._detectors

    @property
    def detectors_high(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.HIGH]

    @property
    def detectors_medium(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.MEDIUM]

    @property
    def detectors_low(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.LOW]

    @property
    def detectors_informational(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.INFORMATIONAL]

    @property
    def detectors_optimization(self):
        return [d for d in self.detectors if d.IMPACT == DetectorClassification.OPTIMIZATION]

    def register_detector(self, detector_class: type[AbstractDetector]) -> None:
        """
        :param detector_class: Class inheriting from `AbstractDetector`.
        """
        _check_common_things("detector", detector_class, AbstractDetector, self._detectors)

        for compilation_unit in self.compilation_units:
            instance = detector_class(compilation_unit, self, logger_detector)
            self._detectors.append(instance)

    def unregister_detector(self, detector_class: type[AbstractDetector]) -> None:
        """
        :param detector_class: Class inheriting from `AbstractDetector`.
        """

        for obj in self._detectors:
            if isinstance(obj, detector_class):
                self._detectors.remove(obj)
                return

    def register_printer(self, printer_class: type[AbstractPrinter]) -> None:
        """
        :param printer_class: Class inheriting from `AbstractPrinter`.
        """
        _check_common_things("printer", printer_class, AbstractPrinter, self._printers)

        instance = printer_class(self, logger_printer)
        self._printers.append(instance)

    def unregister_printer(self, printer_class: type[AbstractPrinter]) -> None:
        """
        :param printer_class: Class inheriting from `AbstractPrinter`.
        """

        for obj in self._printers:
            if isinstance(obj, printer_class):
                self._printers.remove(obj)
                return

    def run_detectors(self) -> list[dict]:
        """
        :return: List of registered detectors results.
        """

        self.load_previous_results()
        results = [d.detect() for d in self._detectors]

        self.write_results_to_hide()
        return results

    def run_printers(self) -> list[Output]:
        """
        :return: List of registered printers outputs.
        """

        return [p.output(self._crytic_compile.target).data for p in self._printers]

    @property
    def triage_mode(self) -> bool:
        return self._triage_mode
